from dataclasses import dataclass
from django.shortcuts import HttpResponse
from django.utils.functional import Promise

from .response import Response


@dataclass(kw_only=True)
class WindowResponseConfig:

    display_header: bool = True

    def dict(self):
        return vars(self)


class WindowResponse(Response):

    base_template = 'ui/layout.html'

    def __init__(
        self, *,
        title: str | Promise,
        context: dict = None,
        overview_template: str = None,
        header_template: str = None,
        panel_template: str = None,
        style_template: str = None,
        script_template: str = None,
        is_centered: bool = False,
        config: WindowResponseConfig = None
    ):
        super().__init__(template=self.base_template, context=context)
        self.title = title
        self.overview_template = overview_template
        self.header_template = header_template
        self.panel_template = panel_template
        self.style_template = style_template
        self.script_template = script_template
        self.is_centered = is_centered
        self.config = config or WindowResponseConfig()

    def _has_panel(self):
        return bool(self.panel_template)

    def get_context(self):
        context = super().get_context()
        if 'has_panel' not in context.keys():
            context.update(has_panel=self._has_panel())
        context.update({
            'title': self.title,
            'overview_template': self.overview_template,
            'header_template': self.header_template,
            'panel_template': self.panel_template,
            'style_template': self.style_template,
            'script_template': self.script_template,
            'is_centered': self.is_centered,
            'config': self.config.dict()
        })
        return context

    def response(self, request, extra_content: str = None, replace_body: bool = True) -> HttpResponse:
        return super().response(request, extra_content, replace_body)



from django.contrib import admin
from .models import Sequence


class SequenceAdmin(admin.ModelAdmin):
    model = Sequence
    list_display = ('name', 'nextval', 'step', 'tenant')
    search_fields = ('name', 'tenant__name')
    list_filter = ['tenant']


admin.site.register(Sequence, SequenceAdmin)



# Generated by Django 5.2.8 on 2026-03-08 08:40

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('log', '0008_activity_log_activity'),
    ]

    operations = [
        migrations.AlterField(
            model_name='activity',
            name='model',
            field=models.CharField(blank=True, help_text='Information about the triggering model, log lines can have deviating modelsCombination of app label and model name separated by a dot.', max_length=255, null=True, verbose_name='Model'),
        ),
        migrations.AlterField(
            model_name='log',
            name='model',
            field=models.CharField(help_text='Combination of app label and model name separated by a dot', max_length=255, verbose_name='Model'),
        ),
        migrations.AlterField(
            model_name='logconfig',
            name='model',
            field=models.CharField(help_text='Combination of app label and model name separated by a dot', max_length=255, verbose_name='Model'),
        ),
    ]



import ast
import json
from dataclasses import dataclass

from django.http import HttpResponse
from django.template.loader import render_to_string


class Response:

    def __init__(self, *, template: str, context: dict):
        self.template = template
        self.context = context

    def add_trigger(self, response):
        pass

    def get_context(self):
        return self.context

    def render(self, request) -> str:
        return render_to_string(
            template_name=self.template, context=self.get_context(), request=request
        )

    def response(self, request, extra_content: str = None, replace_body: bool = False) -> HttpResponse:
        extra_content = extra_content or ''
        res = HttpResponse(content=(
            self.render(request)
            + render_to_string('ui/message.html', request=request)
            + extra_content
        ))
        self.add_trigger(res)
        if replace_body:
            res.headers['HX-Retarget'] = 'body'
            res.headers['HX-Reswap'] = 'innerHTML'
            res.headers['HX-Push-Url'] = request.path
        return res


class OobResponse(Response):

    oob_template = 'ui/oob.html'

    def __init__(
        self, *,
        template: str,
        context: dict,
        swap: str = '',
        tag: str = 'div',
        wrap: bool = True,
        attrs: dict = None
    ):
        super().__init__(template=self.oob_template, context=context)
        self.include_template = template
        self.swap = swap
        self.tag = tag
        self.wrap = wrap
        self.attrs = attrs or {}

    def get_context(self):
        context = super().get_context()
        context.update({'oob': {
            'template': self.include_template,
            'swap': self.swap,
            'tag': self.tag,
            'wrap': self.wrap,
            'attrs': self.attrs
        }})
        return context


@dataclass
class ClientTrigger:

    trigger: dict | str
    header: str = 'HX-Trigger'


class TriggerResponse:

    def __init__(self, trigger: list[ClientTrigger]):
        self.trigger = trigger

    def response(self):
        res = HttpResponse()
        res.headers['HX-Reswap'] = 'none'
        for trigger in self.trigger:
            add_trigger(res, trigger.trigger, trigger.header)
        return res


def search_select_response(queryset) -> HttpResponse:
    return HttpResponse(render_to_string(
        'ui/widgets/model_search_select_options.html',
        {'options': queryset}
    ))


def message_response(request, persistent: bool = False):
    return HttpResponse(content=(render_to_string(
        'ui/message.html', context={'persistent': persistent}, request=request
    )))


def update(request, responses: list[Response], trigger: list[ClientTrigger] = None) -> HttpResponse:
    response = HttpResponse()
    content = ''
    for res in responses:
        content += res.render(request)
        res.add_trigger(response)
    content += render_to_string('ui/message.html', request=request)
    response.content = content
    for t in trigger or []:
        add_trigger(response, t.trigger, t.header)
    return response


def redirect_response(url: str, reload: bool = True):
    res = HttpResponse()
    header = 'HX-Redirect' if reload else 'HX-Location'
    res[header] = url
    return res


def add_trigger(
    response: HttpResponse,
    trigger: dict | str,
    header: str = 'HX-Trigger'
) -> HttpResponse:
    if isinstance(trigger, str):
        trigger = {trigger: ''}
    res_trigger = response.headers.get(header)
    if not res_trigger:
        response.headers[header] = json.dumps(trigger)
        return response
    try:
        res_trigger = ast.literal_eval(response.headers.get(header, '{}'))
    except SyntaxError:
        res_trigger = {response.headers[header]: ''}
    res_trigger.update(trigger)
    response.headers[header] = json.dumps(res_trigger)
    return response



from django.utils import translation
from django.test import TestCase
from accrete.models import AccessGroup
from accrete.utils import translated_db_value


class FieldTestCase(TestCase):

    lang_settings = {
        'LANGUAGE_CODE': 'en-us',
        'LANGUAGES': [
            ('de-de', 'German'),
            ('en-us', 'English'),
        ]
    }

    def test_assign_string(self):
        with self.settings(**self.lang_settings):
            group = AccessGroup(
                name='Test en-us',
                description='Test Description en-us',
                code='test',
            )
            group.save()
            self.assertEqual(group.name, 'Test en-us')
            self.assertEqual(group.description, 'Test Description en-us')
            translation.activate('de-de')
            self.assertEqual(group.name, 'Test en-us')
            self.assertEqual(group.description, 'Test Description en-us')
            translation.activate('en-us')

    def test_assign_string_german(self):
        with self.settings(**self.lang_settings):
            translation.activate('de-de')
            group = AccessGroup(
                name='Test de-de',
                description='Test Description de-de',
                code='test',
            )
            group.save()
            self.assertEqual(group.name, 'Test de-de')
            self.assertEqual(group.description, 'Test Description de-de')
            translation.activate('en-us')
            self.assertEqual(group.name, 'Test de-de')
            self.assertEqual(group.description, 'Test Description de-de')
            translation.activate('en-us')

    def test_write_string(self):
        with self.settings(**self.lang_settings):
            translation.activate('en-us')
            group = AccessGroup(
                name='Test en-us',
                description='Test Description en-us',
                code='test',
            )
            group.save()
            translation.activate('de-de')
            group.name = 'Test de-de'
            group.description = 'Test Description de-de'
            group.save()
            self.assertEqual(group.name, 'Test de-de')
            self.assertEqual(group.description, 'Test Description de-de')
            translation.activate('en-us')
            group.refresh_from_db()
            self.assertEqual(group.name, 'Test en-us')
            self.assertEqual(group.description, 'Test Description en-us')
            self.assertDictEqual(
                translated_db_value(group, 'name'),
                {'en-us': 'Test en-us', 'de-de': 'Test de-de'}
            )

    def test_assign_dict(self):
        with self.settings(**self.lang_settings):
            group = AccessGroup(
                name={'en-us': 'Test en-us', 'de-de': 'Test de-de'},
                description={
                    'en-us': 'Test Description en-us',
                    'de-de': 'Test Description de-de'
                },
                code='test',
            )
            group.save()
            self.assertEqual(group.name, 'Test en-us')
            self.assertEqual(group.description, 'Test Description en-us')
            translation.activate('de-de')
            group.refresh_from_db()
            self.assertEqual(group.name, 'Test de-de')
            self.assertEqual(group.description, 'Test Description de-de')
            translation.activate('en-us')
            self.assertDictEqual(
                translated_db_value(group, 'name'),
                {'en-us': 'Test en-us', 'de-de': 'Test de-de'}
            )
            self.assertDictEqual(
                translated_db_value(group, 'description'),
                {'en-us': 'Test Description en-us', 'de-de': 'Test Description de-de'}
            )



from django.views.generic import View
from accrete.mixins import TenantRequiredMixin
from .helper import (
    is_htmx,
    get_group_data,
    access_denied_modal_response,
    access_denied_page_response
)


class TenantView(TenantRequiredMixin, View):

    """
    Base View that handles displaying access denied messages
    to the user if member/tenant groups are missing.
    """

    def handle_tenant_group_not_set(self):
        if is_htmx(self.request):
            return access_denied_modal_response(
                self.request,
                get_group_data(self.TENANT_GROUPS, self.MEMBER_GROUPS)
            )
        return access_denied_page_response(
            self.request,
            get_group_data(self.TENANT_GROUPS, self.MEMBER_GROUPS)
        )

    def handle_member_group_not_set(self):
        if is_htmx(self.request):
            return access_denied_modal_response(
                self.request,
                get_group_data(self.TENANT_GROUPS, self.MEMBER_GROUPS)
            )
        return access_denied_page_response(
            self.request,
            get_group_data(self.TENANT_GROUPS, self.MEMBER_GROUPS)
        )



from django.test import TestCase
from django.contrib.auth import get_user_model
from django.test import Client
from accrete import models

User = get_user_model()


class MiddlewareTestCase(TestCase):

    def setUp(self):
        self.user = User(email='test@user.test', username='Test User')
        self.user.save()

    def test_no_tenant(self):
        c = Client()
        c.force_login(self.user)
        res = c.get('/')
        self.assertEqual(res.headers['X-TENANT-ID'], '0')
        self.assertFalse(res.wsgi_request.tenant)

    def test_get_non_existing(self):
        c = Client()
        c.force_login(self.user)
        res = c.get('/', {'tenant_id': 1})
        self.assertEqual(res.headers.get('X-TENANT-ID', '0'), '0')
        self.assertFalse(res.wsgi_request.tenant)
    
    def test_get_no_member(self):
        tenant = models.Tenant(name='Test Tenant')
        tenant.save()
        c = Client()
        c.force_login(self.user)
        res = c.get('/', {'tenant_id': tenant.pk + 1})
        self.assertEqual(res.headers['X-TENANT-ID'], '0')
        self.assertFalse(res.wsgi_request.tenant)
        self.assertFalse(res.wsgi_request.member)

    def test_get(self):
        tenant = models.Tenant(name='Test Tenant')
        tenant.save()
        member = models.Member(
            tenant=tenant,
            user=self.user
        )
        member.save()
        c = Client()
        c.force_login(self.user)
        res = c.get('/')
        self.assertEqual(res.headers['X-TENANT-ID'], str(tenant.pk))
        self.assertEqual(res.wsgi_request.tenant, tenant)
        self.assertEqual(res.wsgi_request.member, member)

    def test_post_no_member(self):
        c = Client()
        c.force_login(self.user)
        res = c.post('/', {'tenant_id': 1})
        self.assertEqual(res.headers.get('X-TENANT-ID', '0'), '0')

    def test_post(self):
        tenant = models.Tenant(name='Test Tenant')
        tenant.save()
        member = models.Member(
            tenant=tenant,
            user=self.user
        )
        member.save()
        c = Client()
        c.force_login(self.user)
        res = c.post('/', {'tenant_id': tenant.pk})
        self.assertEqual(res.headers['X-TENANT-ID'], str(tenant.pk))



import json
import contextvars

from django.db import models
from django.db.models.query import RawQuerySet
from django.apps import apps

from accrete.fields import TranslatedCharField, TranslatedTextField
from accrete.utils.models import model_to_dict
from accrete.tenant import get_tenant, get_member

activity_contextvar = contextvars.ContextVar('activity', default=None)

TYPES_FK = ['AutoField', 'BigAutoField', 'ForeignKey']
TYPES_INT = ['IntegerField', 'PositiveSmallIntegerField']
TYPES_DECIMAL = ['DecimalField']
TYPES_FLOAT = ['FloatField']
TYPES_STR = ['CharField', 'TextField']
TYPES_BOOL = ['BooleanField']
TYPES_DATETIME = ['DateTimeField']
TYPES_DATE = ['DateField']
TYPES_TIME = ['TimeField']
TYPES_JSON = ['JSONField']


class ActivityContext:

    def __init__(
        self,
        code: str, *,
        description: str | dict = None,
        instance: models.Model = None
    ):
        self.code = code
        self.description = description
        self.instance = instance
        tenant = get_tenant() or getattr(instance, 'tenant', None)
        member = get_member()
        Activity = apps.get_model('log', 'Activity')
        self.activity = Activity(
            code=code,
            description=self.description,
        )
        if tenant:
            self.activity.tenant = tenant
        if member:
            self.activity.user = member and member.user
        self.set_instance(self.instance)
        self.activity.save()

    def __enter__(self):
        set_activity(self.activity)
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        set_activity(None)

    def set_instance(self, instance):
        self.instance = instance
        self.activity.model = (
            self.instance
            and f'{self.instance._meta.app_label}.{self.instance._meta.model_name}'
            or None
        )
        self.activity.object_id = self.instance and self.instance.pk or None
        self.activity.save()


def set_activity(activity_instance):
    activity_contextvar.set(activity_instance)


def get_activity():
    return activity_contextvar.get()


def activity(
    code: str, *,
    description: str | dict = None,
    instance: models.Model = None
):
    return ActivityContext(
        code=code,
        description=description,
        instance=instance
    )


def internal_type_to_log_type(field: models.Field):
    internal_type = field.get_internal_type()
    if internal_type in TYPES_FK:
        return 'fk'
    if internal_type in TYPES_INT:
        return 'int'
    if internal_type in TYPES_FLOAT:
        return 'float'
    if internal_type in TYPES_DECIMAL:
        return 'decimal'
    if internal_type in TYPES_BOOL:
        return 'bool'
    if internal_type in TYPES_STR:
        return 'str'
    if internal_type in TYPES_DATE:
        return 'date'
    if internal_type in TYPES_DATETIME:
        return 'datetime'
    if internal_type in TYPES_JSON:
        return 'json'


def log_state_to_dict(logs: RawQuerySet) -> tuple[dict, dict]:
    state = dict()
    info = dict()
    if not logs:
        return state, info
    for log in logs:
        state.update({log.field: log.cast_value()})
        info.update({log.field: {'new_value_type': log.new_value_type}})
    return state, info

def get_instance_state(instance: models.Model, log_config: models.Model = None):
    if not log_config:
        LogConfig = apps.get_model('log', 'LogConfig')
        log_config = LogConfig.objects.filter(
            model__iexact=f'{instance._meta.app_label}.{instance._meta.model_name}'
        ).prefetch_related('fields').first()
    cleaned_state = dict()
    if not log_config:
        return cleaned_state
    all_fields = list(map(lambda x: x.name, instance._meta.get_fields()))
    if log_config.exclude_fields:
        fields_to_log = list(set(all_fields).difference(
            set(log_config.fields.all().values_list('field_name', flat=True))
        ))
    else:
        fields_to_log = log_config.fields.all().values_list(
            'field_name', flat=True
        )
    state = model_to_dict(instance, translated=False)
    for f, v in state.items():
        if isinstance(v, (list, tuple)):
            continue
        if f not in fields_to_log:
            continue
        if (
            isinstance(v, dict)
            and isinstance(
                instance._meta.get_field(f), 
                (TranslatedCharField, TranslatedTextField),
            )
        ):
            cleaned_state.update({f: json.dumps(v)})
            continue
        cleaned_state.update({f: v})
    return cleaned_state



from uuid import uuid4
import logging

from django.contrib.auth import get_user_model
from django.utils.http import urlsafe_base64_encode
from django.core.validators import validate_email
from django.utils.translation import gettext_lazy as _
from django import forms

from .models import UserConfirmationToken

_logger = logging.getLogger(__name__)
User = get_user_model()


class UserRegistrationForm(forms.Form):
    email = forms.EmailField(
        label='E-Mail', widget=forms.EmailInput
    )

    password1 = forms.CharField(
        label=_('Password'),
        widget=forms.PasswordInput,
        max_length=128
    )

    password2 = forms.CharField(
        label=_('Confirm Password'),
        widget=forms.PasswordInput,
        max_length=128
    )

    def clean_password1(self):
        pw = self.cleaned_data['password1']
        if len(pw) < 1:
            raise forms.ValidationError(_(
                'Password must be at least 8 characters long!'
            ))
        return pw

    def clean(self):
        super().clean()
        password1 = self.cleaned_data.get("password1")
        password2 = self.cleaned_data.get("password2")
        msg = _("Passwords don't match!")

        if password1 and password2 and password1 != password2:
            self.add_error('password2', msg)

        return self.cleaned_data

    def clean_email(self):
        email = self.cleaned_data.get('email', '').lower()
        validate_email(email)

        if User.objects.filter(email=email):
            raise forms.ValidationError(
                _('This email address is already bound to an account.')
            )

        return email

    def save(self):
        user = User(
            email=self.cleaned_data.get('email'),
            is_staff=False,
            is_superuser=False,
            is_active=False
        )
        user.set_password(self.cleaned_data["password1"])
        user.save()
        _logger.info(f'Registering user {user}')

        token = urlsafe_base64_encode(str(uuid4()).encode())
        user_confirmation_token = UserConfirmationToken(
            user=user,
            token=token
        )
        user_confirmation_token.save()

        return user


class ResendConfirmationMailForm(forms.Form):
    email = forms.EmailField(
        label='E-Mail', widget=forms.EmailInput
    )

    def clean_email(self):
        email = self.cleaned_data['email']
        validate_email(email)
        return email


class ConfirmRegistrationForm(forms.Form):
    token = forms.CharField()

    def clean_token(self):
        token = self.cleaned_data['token']
        user_token = self.get_user_token(token)
        if not token or not user_token:
            raise forms.ValidationError(_(
                'Invalid token supplied.'
            ))

        if user_token.user.is_active:
            raise forms.ValidationError(_(
                'User already activated'
            ))

        return token

    @staticmethod
    def get_user_token(token):
        return UserConfirmationToken.objects.filter(token=token).first()

    def save(self):
        user_token = self.get_user_token(self.cleaned_data['token'])
        user_token.user.is_active = True
        user_token.user.save()
        user_token.delete()
        _logger.info(f'Confirmed User Registration for {user_token.user}')



# Generated by Django 5.2.7 on 2025-10-20 17:31

import accrete.fields
import django.db.models.deletion
from django.conf import settings
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('accrete', '0010_alter_accessgroup_description'),
        ('log', '0007_alter_log_new_value_type_alter_log_old_value_type'),
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name='Activity',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('code', models.CharField(blank=True, default='', verbose_name='Code')),
                ('description', accrete.fields.TranslatedCharField(blank=True, null=True, verbose_name='Description')),
                ('model', models.CharField(blank=True, help_text='Combination of app label and model name seperated by a dot', max_length=255, null=True, verbose_name='Model')),
                ('object_id', models.BigIntegerField(blank=True, null=True, verbose_name='Object ID')),
                ('log_date', models.DateTimeField(auto_now_add=True, verbose_name='Date')),
                ('tenant', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, related_name='accrete_activities', to='accrete.tenant', verbose_name='Tenant')),
                ('user', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='accrete_activities', to=settings.AUTH_USER_MODEL)),
            ],
            options={
                'verbose_name': 'Activity',
                'verbose_name_plural': 'Activities',
                'db_table': 'accrete_activity',
            },
        ),
        migrations.AddField(
            model_name='log',
            name='activity',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, related_name='logs', to='log.activity'),
        ),
    ]




from django.contrib.auth.mixins import LoginRequiredMixin
from django.core.exceptions import ImproperlyConfigured
from django.shortcuts import redirect, resolve_url
from accrete.tenant import get_tenant, tenant_has_group, member_has_group
from accrete import config


class TenantRequiredMixin(LoginRequiredMixin):

    # Redirect to the specified url if the group check fails
    TENANT_NOT_SET_URL = None
    GROUP_NOT_SET_URL = None

    # If set, one of the supplied groups must be set on the
    # tenant and member respectively. If the list item is a tuple,
    # all the groups in the tuple must be set.
    TENANT_GROUPS: list[str | tuple[str]] = []
    MEMBER_GROUPS: list[str | tuple[str]] = []
    STAFF_ACCESS_GROUP_BYPASS: bool = True

    def dispatch(self, request, *args, **kwargs):
        if not self.get_tenant():
            return self.handle_tenant_not_set()
        if self.STAFF_ACCESS_GROUP_BYPASS and request.user and request.user.is_staff:
            return super().dispatch(request, *args, **kwargs)
        if not self.check_tenant_group():
            return self.handle_tenant_group_not_set()
        if not self.check_member_group():
            return self.handle_member_group_not_set()
        return super().dispatch(request, *args, **kwargs)

    def handle_tenant_not_set(self):
        return redirect(
            resolve_url(self.get_tenant_not_set_url())
            + f'?next={self.request.get_full_path_info()}'
        )

    def handle_tenant_group_not_set(self):
        return redirect(self.get_group_not_set_url())

    def handle_member_group_not_set(self):
        return redirect(self.get_group_not_set_url())

    def get_tenant_not_set_url(self):
        tenant_not_set_url = (
            self.TENANT_NOT_SET_URL
            or config.ACCRETE_TENANT_NOT_SET_URL
        )
        if not tenant_not_set_url:
            cls_name = self.__class__.__name__
            raise ImproperlyConfigured(
                f"{cls_name} is missing the tenant_not_set_url attribute. "
                f"Define {cls_name}.TENANT_NOT_SET_URL, "
                f"settings.ACCRETE_TENANT_NOT_SET_URL, or override "
                f"{cls_name}.get_tenant_not_set_url()."
            )
        return tenant_not_set_url

    def get_group_not_set_url(self):
        group_not_set_url = (
            self.GROUP_NOT_SET_URL
            or config.ACCRETE_GROUP_NOT_SET_URL
        )
        if not group_not_set_url:
            cls_name = self.__class__.__name__
            raise ImproperlyConfigured(
                f"{cls_name} is missing the group_not_set_url attribute. "
                f"Define {cls_name}.GROUP_NOT_SET_URL, "
                f"settings.ACCRETE_GROUP_NOT_SET_URL, or override "
                f"{cls_name}.get_group_not_set_url()."
            )
        return group_not_set_url

    def check_tenant_group(self) -> bool:
        if not self.TENANT_GROUPS:
            return True
        for group in self.TENANT_GROUPS:
            if isinstance(group, tuple) and all([tenant_has_group(g) for g in group]):
                return True
            elif tenant_has_group(group):
                return True
        return False

    def check_member_group(self) -> bool:
        if not self.MEMBER_GROUPS:
            return True
        for group in self.MEMBER_GROUPS:
            if isinstance(group, tuple) and all([member_has_group(g) for g in group]):
                return True
            elif member_has_group(group):
                return True
        return False

    @staticmethod
    def get_tenant():
        return get_tenant()



from django.test import TestCase
from django.contrib.auth import get_user_model
from django.forms import ModelChoiceField

from accrete import models
from accrete import forms
from accrete.tenant import set_tenant

User = get_user_model()


class TestForm(forms.TenantForm):

    members = ModelChoiceField(
        queryset=models.Member.objects.all()
    )


class FormTestCase(TestCase):

    def setUp(self):
        self.tenant = models.Tenant(name='Test Tenant')
        self.tenant.save()
        self.user = User(email='test@user.test', username='Test User')
        self.user.save()

    def test_modelform_set_tenant(self):
        set_tenant(self.tenant)
        f = forms.TenantForm()
        self.assertEqual(self.tenant, f.tenant)
        set_tenant(None)

    def test_modelform_tenant_queryset(self):
        set_tenant(self.tenant)
        f = TestForm()
        self.assertQuerySetEqual(
            f.fields['members'].queryset,
            models.Member.objects.filter(tenant=self.tenant)
        )
        set_tenant(None)

    def test_modelform_clean_staff(self):
        self.user.is_staff = True
        self.user.save()
        tenant = models.Tenant(name='Test 1')
        tenant.save()
        tenant_2 = models.Tenant(name='Test 2')
        tenant_2.save()
        member = models.Member(user=self.user, tenant=tenant)
        member.save()
        member_2 = models.Member(user=self.user, tenant=tenant_2)
        member_2.save()
        set_tenant(tenant)
        f = TestForm()
        self.assertQuerySetEqual(f.fields['members'].queryset, models.Member.objects.filter(pk=member.pk))
        set_tenant(None)

    def test_modelform_no_tenant(self):
        set_tenant(None)
        f = TestForm()
        self.assertQuerySetEqual(
            f.fields['members'].queryset,
            models.Member.objects.none()
        )

    def test_modelform_with_member(self):
        member = models.Member(
            tenant=self.tenant,
            user=self.user
        )
        member.save()
        set_tenant(self.tenant)
        f = forms.TenantForm({})
        self.assertTrue(f.is_valid())
        set_tenant(None)



from django import forms
from django.utils.translation import gettext_lazy as _
from . models import Theme


class ThemeForm(forms.ModelForm):

    class Meta:
        model = Theme
        fields = [
            'user',
            'tenant',
            'base_theme',
            'color_primary',
            'color_success',
            'color_link',
            'color_warning',
            'color_danger',
            'navbar_color',
        ]
        widgets = {
            'color_primary': forms.widgets.TextInput(
                attrs={'type': 'color', 'autocomplete': 'off'}
            ),
            'color_success': forms.widgets.TextInput(
                attrs={'type': 'color', 'autocomplete': 'off'}
            ),
            'color_link': forms.widgets.TextInput(
                attrs={'type': 'color', 'autocomplete': 'off'}
            ),
            'color_warning': forms.widgets.TextInput(
                attrs={'type': 'color', 'autocomplete': 'off'}
            ),
            'color_danger': forms.widgets.TextInput(
                attrs={'type': 'color', 'autocomplete': 'off'}
            ),
            'user': forms.widgets.HiddenInput(),
            'tenant': forms.widgets.HiddenInput()
        }

    def _clean_color(self, hex_color: str):
        if not hex_color.startswith('#'):
            hex_color = f'#{hex_color}'
        if len(hex_color) != 7:
            raise forms.ValidationError(_(
                'Color must be of length 7 and including the starting "#"'
            ))
        return hex_color

    def clean_color_primary(self):
        color = self._clean_color(self.cleaned_data.get('color_primary'))
        return color

    def clean_color_success(self):
        color = self._clean_color(self.cleaned_data.get('color_success'))
        return color

    def clean_color_link(self):
        color = self._clean_color(self.cleaned_data.get('color_link'))
        return color

    def clean_color_warning(self):
        color = self._clean_color(self.cleaned_data.get('color_warning'))
        return color

    def clean_color_danger(self):
        color = self._clean_color(self.cleaned_data.get('color_danger'))
        return color



# Generated by Django 5.2.7 on 2025-10-05 21:45

from django.conf import settings
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('accrete', '0010_alter_accessgroup_description'),
        ('log', '0005_logconfig_exclude_fields_and_more'),
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.RemoveIndex(
            model_name='log',
            name='accrete_log_model_cf888c_idx',
        ),
        migrations.AddIndex(
            model_name='log',
            index=models.Index(fields=['model', 'object_id', 'tenant_id'], name='accrete_log_model_22d992_idx'),
        ),
    ]



# Generated by Django 5.2.1 on 2025-08-08 06:13

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('user', '0009_alter_user_theme'),
    ]

    operations = [
        migrations.AlterField(
            model_name='user',
            name='theme',
            field=models.CharField(choices=[('preset', 'Preset'), ('light', 'Light'), ('dark', 'Dark'), ('custom', 'Custom')], default='preset', max_length=50, verbose_name='Theme'),
        ),
    ]



# Generated by Django 5.2.5 on 2025-10-03 07:29

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('ui', '0004_theme_force_tenant_theme'),
    ]

    operations = [
        migrations.AddField(
            model_name='theme',
            name='base_theme',
            field=models.CharField(choices=[('light', 'Light'), ('dark', 'Dark')], default='light', verbose_name='Base Theme'),
        ),
    ]



from datetime import datetime, date, time
from dataclasses import dataclass
from django.db import transaction
from django.db.models import F
from django.utils import timezone
from django.apps import apps

from accrete.tenant import get_tenant


@dataclass
class SequenceData:

    counter: int
    number: str


def get_nextval(
    name: str, 
    per_date: datetime | date = None, 
    create_values: dict = None
) -> SequenceData:
    Sequence = apps.get_model('sequence', 'Sequence')
    tenant = get_tenant()
    sequence = Sequence.objects.filter(tenant_id=tenant.pk, name=name).first()
    if not sequence:
        create_values = create_values or {}
        create_values.update(name=name, tenant_id=tenant.pk)
        sequence = Sequence(**create_values)
        sequence.save()
    if isinstance(per_date, date):
        per_date = datetime.combine(per_date, time.min)
    per_date = per_date or timezone.now()
    year = per_date.year
    if sequence.has_subsequence_per_year:
        name = f'{name}__{year}'
        sequence = Sequence.objects.get_or_create(
            defaults={
                'name': name,
                'has_subsequence_per_year': False,
                'step': sequence.step,
                'prefix': sequence.prefix,
                'padding': sequence.padding,
            },
            name=name,
            tenant_id=tenant.pk,
        )[0]
    counter = _get_nextval(name)
    placeholder = {
        'Y': year,
        'y': str(year)[2:4],
        'm': str(per_date.month).zfill(2),
        'd': str(per_date.day).zfill(2),
        'H': per_date.hour,
        'M': per_date.minute,
        'S': per_date.second,
    }
    number = sequence.prefix % placeholder
    number += str(counter).zfill(sequence.padding)
    return SequenceData(
        counter=counter,
        number=number,
    )


def _get_nextval(name: str, create_values: dict = None) -> int:
    Sequence = apps.get_model('sequence', 'Sequence')
    tenant = get_tenant()
    with transaction.atomic():
        seq = Sequence.objects.filter(
            tenant_id=tenant.pk, name=name
        ).select_for_update().first()

        if seq is None and create_values is None:
            raise ValueError(f'Sequence "{name}" does not exist.')
        elif seq is None:
            create_values = create_values or {}
            create_values.update(name=name, tenant_id=tenant.pk)
            seq = Sequence(**create_values)
            seq.save()

        nextval = seq.nextval
        seq.nextval = F('nextval') + seq.step
        seq.save()

    return nextval



import logging
from django.conf import settings
from django.utils.translation import gettext_lazy as _

_logger = logging.getLogger(__name__)

ACCRETE_USER_REGISTRATION_MAIL_FROM_NAME = getattr(
    settings, 'ACCRETE_USER_REGISTRATION_MAIL_FROM_NAME',
    False
)

ACCRETE_USER_REGISTRATION_TEMPLATE_NAME = getattr(
    settings, 'ACCRETE_USER_REGISTRATION_TEMPLATE_NAME',
    'user_registration/mail_templates/confirmation_mail.html'
)

ACCRETE_USER_REGISTRATION_MAIL_SUBJECT = getattr(
    settings, 'ACCRETE_USER_REGISTRATION_MAIL_SUBJECT',
    _('Registration Confirmation')
)

ACCRETE_USER_REGISTRATION_ALLOWED = getattr(
    settings, 'ACCRETE_USER_REGISTRATION_ALLOWED', True
)

if not ACCRETE_USER_REGISTRATION_MAIL_FROM_NAME:
    _logger.warning(
        'Setting "ACCRETE_USER_REGISTRATION_MAIL_FROM_NAME" missing.\n'
        'User Registration won\'t work. Set it or remove '
        'the app "accrete.contrib.user_registration".'
    )



# Generated by Django 5.2.8 on 2026-03-08 08:40

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('ui', '0005_theme_base_theme'),
    ]

    operations = [
        migrations.AddField(
            model_name='theme',
            name='navbar_color',
            field=models.CharField(blank=True, choices=[('is-primary', 'Primary'), ('is-success', 'Success'), ('is-link', 'Link'), ('is-warning', 'Warning'), ('is-danger', 'Danger')], default='is-primary', null=True, verbose_name='Navbar Color'),
        ),
    ]



# Generated by Django 5.1.7 on 2025-03-23 08:11

import accrete.contrib.user.models
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('user', '0007_user_managed_login'),
    ]

    operations = [
        migrations.RemoveConstraint(
            model_name='user',
            name='no_email_for_managed_user',
        ),
        migrations.AlterField(
            model_name='user',
            name='language_code',
            field=models.CharField(blank=True, default=accrete.contrib.user.models.default_language_code, max_length=10, null=True, verbose_name='Language'),
        ),
    ]



import json
from django.test import TestCase
from django.utils import translation
from django.conf import settings
from accrete.models import AccessGroup, Tenant
from accrete.tenant import set_tenant
from accrete.contrib.log import activity
from accrete.contrib.log.models import Activity, Log, LogConfig, LogConfigField
from accrete.contrib.log import helper, queries


class LogTestCase(TestCase):

    def setUp(self):
        super().setUp()
        self.tenant = Tenant(name='Test Tenant', is_active=True)
        self.tenant.save()
        set_tenant(self.tenant)
        self.log_config = LogConfig(
            model='accrete.AccessGroup', 
            exclude_fields=True, 
            ignore_errors=False,
        )
        self.log_config.save()

    def test_activity_context(self):
        group = AccessGroup(code='test', name='Test Group')
        with activity(
            code='test context', 
            description='Test description',
        ) as context:
            group.save()
            self.assertEqual(context.activity.description, 'Test description')
            self.assertTrue(context.activity.logs.exists())
    
    def test_log_specific_fields(self):
        self.log_config.exclude_fields = False
        self.log_config.save()
        LogConfigField.objects.bulk_create([
            LogConfigField(log_config=self.log_config, field_name='name'),
            LogConfigField(log_config=self.log_config, field_name='apply_on'),
        ])
        group = AccessGroup(code='test specific field log')
        with activity(code='test_log_specific') as ctx:
            group.save()
            self.assertEqual(ctx.activity.logs.count(), 2)
        self.log_config.fields.all().delete()
        self.log_config.exclude_fields = True
        self.log_config.save()


    
    def test_instance_state(self):
        translation.activate(settings.LANGUAGE_CODE)
        group = AccessGroup(code='test state', description='Group Description')
        group.save()
        state = helper.get_instance_state(group)
        expected_state = {
            'id': group.pk,
            'code': 'test state',
            'name': json.dumps({settings.LANGUAGE_CODE: ''}),
            'description': json.dumps({settings.LANGUAGE_CODE: 'Group Description'}),
            'apply_on': 'tenant',
        }
        self.assertDictEqual(state, expected_state)
        group.apply_on = 'member'
        group.name = 'Test'
        group.save()
        expected_state.update(
            apply_on='member',
            name=json.dumps({settings.LANGUAGE_CODE: 'Test'})
        )
        state = helper.get_instance_state(group)
        self.assertDictEqual(state, expected_state)
    
    def test_log_state(self):
        translation.activate(settings.LANGUAGE_CODE)
        group = AccessGroup(code='test log state')
        group.save()
        group.description = 'Description'
        group.save()
        log_state = queries.current_log_state('accrete.AccessGroup', group.pk)
        expected = {
            'id': group.pk,
            'code': 'test log state',
            'name': json.dumps({settings.LANGUAGE_CODE: ''}),
            'description': json.dumps({settings.LANGUAGE_CODE: 'Description'}),
            'apply_on': 'tenant',
        }
        self.assertDictEqual(helper.log_state_to_dict(log_state)[0], expected)



from channels.generic.websocket import WebsocketConsumer, JsonWebsocketConsumer

from accrete.tenant import get_tenant


class WebsocketTenantConsumer(WebsocketConsumer):

    def websocket_connect(self, message):
        tenant = get_tenant()
        if tenant is None:
            raise ValueError('Tenant must be set.')
        super().websocket_connect(message)


class JsonWebsocketTenantConsumer(JsonWebsocketConsumer):

    def websocket_connect(self, message):
        tenant = get_tenant()
        if tenant is None:
            raise ValueError('Tenant must be set.')
        super().websocket_connect(message)



from django.contrib import admin
from . import models


class LogConfigFieldInLine(admin.TabularInline):

    model = models.LogConfigField


class LogConfigAdmin(admin.ModelAdmin):

    model = models.LogConfig
    list_display = ('model', 'ignore_errors', 'exclude_fields')
    search_fields = ('pk', 'model')
    list_filter = ['ignore_errors', 'exclude_fields']
    inlines = [LogConfigFieldInLine]


class LogAdmin(admin.ModelAdmin):

    model = models.LogConfig
    list_display = (
        'model', 'object_id', 'log_date', 'field', 'old_value', 'new_value',
        'activity', 'user', 'tenant'
    )
    search_fields = ('model', 'field', 'object_id', 'old_value', 'new_value')
    list_filter = ['model', 'tenant']


class ActivityLogInline(admin.TabularInline):

    model = models.Log
    extra = 0


class ActivityAdmin(admin.ModelAdmin):

    model = models.Activity
    list_display = ('code', 'description', 'model', 'object_id', 'user', 'tenant')
    search_fields = ['code', 'description', 'model', 'object_id']
    list_filter = ['code', 'tenant']
    inlines = [ActivityLogInline]


admin.site.register(models.LogConfig, LogConfigAdmin)
admin.site.register(models.Log, LogAdmin)
admin.site.register(models.Activity, ActivityAdmin)



import json

from django.db import connection
from django.db.models import Model, Field
from django.conf import settings
from django.forms import model_to_dict as mtd
from django.utils import translation
from django.utils.functional import Promise

from accrete.fields import TranslatedCharField, TranslatedTextField


def get_related_model(model: type[Model], rel_path: str) -> tuple[Model, list[str]]:
    names = []
    related_model = model
    for part in rel_path.split('__'):
        names.append(str(related_model._meta.verbose_name))
        try:
            next_model = related_model._meta.fields_map[part].related_model
            related_model = next_model
        except (AttributeError, KeyError):
            try:
                next_model = getattr(related_model, part).field.related_model
                if next_model is not None:
                    related_model = next_model
            except AttributeError:
                break
    names.append(str(related_model._meta.verbose_name))
    return related_model, names


def get_related_field(model: type[Model], field_path: str) -> Field:
    parts = field_path.split('__')
    rel_path = '__'.join(parts[:-1])
    rel_model, _ = get_related_model(model, rel_path)
    return rel_model._meta.get_field(parts[-1])


def translated_db_value(instance: Model, field: str) -> dict:
    if not instance.pk:
        return dict()
    with connection.cursor() as cr:
        query = """SELECT %s FROM %s WHERE id = %s""" % (
            field,
            instance._meta.db_table,
            instance.pk
        )
        cr.execute(query)
        row = cr.fetchone()
        row = row and row[0]
        db_value = row and json.loads(row) or dict()
    return db_value


def translate_all(text: Promise, values: dict, languages: list = None) -> dict:
    translations = {}
    initial_language = translation.get_language()
    languages = languages or [lang[0] for lang in (settings.LANGUAGES or [])]
    for language in languages:
        translation.activate(language)
        translations.update({language: str(text % values)})
    translation.activate(initial_language)
    return translations


def model_to_dict(instance, fields=None, exclude=None, translated=True):
    """
    Extends django.forms.model_to_dict with the option to get the
    translation dict instead of the translated string for the active language
    """
    data = mtd(instance, fields=fields, exclude=exclude)
    if translated:
        return data
    updates = {}
    for f, v in data.items():
        if isinstance(instance._meta.get_field(f), (TranslatedCharField, TranslatedTextField)):
            updates.update({f: translated_db_value(instance, f)})
    data.update(updates)
    return data



from django.core import paginator
from django.db import models
from django.http import HttpResponse
from django.utils.translation import gettext_lazy as _
from django.utils.functional import Promise
from django.utils.safestring import mark_safe
from accrete.contrib.ui import Filter
from .response import Response, OobResponse
from .window import WindowResponse, WindowResponseConfig


class TableResponse(WindowResponse):

    def __init__(
        self, *,
        title: str | Promise,
        context: dict,
        fields: list[str],
        base_template: str = None,
        instance_label: str | Promise | None = None,
        active_instance: models.Model = None,
        footer: dict = None,
        page: paginator.Page = None,
        ui_filter: Filter = None,
        endless_scroll: bool = False,
        header_template: str = None,
        panel_template: str = None,
        overview_template: str = 'ui/table.html',
        detail_header_template: str = None,
        detail_data_template: str = None,
        detail_enabled: bool = True,
        can_compact: bool = True,
        config: WindowResponseConfig = None
    ):
        assert page is not None or ui_filter is not None, _(
            'Argument page or ui_filter must be supplied'
        )
        super().__init__(
            title=title,
            context=context,
            overview_template=overview_template,
            header_template=header_template,
            panel_template=panel_template,
            config=config
        )
        self.instance_label = instance_label
        self.active_instance = active_instance
        self.fields = fields
        self.base_template = base_template or self.base_template
        self.footer = footer or {}
        self.page = page or (ui_filter and ui_filter.get_page())
        self.ui_filter = ui_filter
        self.endless_scroll = endless_scroll
        self.detail_header_template = detail_header_template
        self.detail_data_template = detail_data_template
        self.detail_enabled = detail_enabled
        self.can_compact = can_compact

    def _has_panel(self):
        return bool(self.panel_template or self.ui_filter)

    def get_context(self):
        context = super().get_context()
        has_url = hasattr(self.page.object_list.first(), 'get_absolute_url')
        context.update({
            'page': self.page,
            'ui_filter': self.ui_filter,
            'endless_scroll': self.endless_scroll,
            'fields': self.fields,
            'instance_label': self.instance_label,
            'active_instance': self.active_instance,
            'table_data': self.table_data(),
            'footer': self.footer,
            'detail_header_template': self.detail_header_template,
            'detail_data_template': self.detail_data_template,
            'detail_enabled': self.detail_enabled,
            'can_compact': self.can_compact,
            'has_url': has_url,
            'show_content_right': str(bool(
                (self.active_instance and has_url and self.detail_enabled)
                or (self.detail_header_template or self.detail_data_template)
            )).lower()
        })
        return context

    def table_data(self):
        data = {
            'header': {'fields': []},
            'body': {'rows': []},
            'footer': {'fields': []},
        }
        meta_data = self.table_meta_data(self.page.object_list.model)
        if self.instance_label:
            data['header']['fields'].insert(0, {
                'content': self.instance_label, 'alignment': 'left'
            })
            data['footer']['fields'].insert(0, {
                'content': '', 'alignment': 'left'})
        for field_name in self.fields:
            alignment = meta_data.get(field_name, {}).get('alignment', 'left')
            verbose_name = meta_data.get(field_name, {}).get('verbose_name', '')
            data['header']['fields'].append({
                'content': verbose_name,
                'alignment': alignment,
            })
            data['footer']['fields'].append({
                'content': self.footer.get(field_name, ''),
                'verbose_name': verbose_name,
                'alignment': alignment,
            })
        for instance in self.page.object_list:
            row_data = {
                'fields': [],
                'id': instance.pk,
                'is_active': bool(instance == self.active_instance),
                'hx_get': self.detail_enabled and self.table_row_url(instance) or False,
            }
            if self.instance_label:
                row_data['fields'].insert(0, {'content': str(instance)})
            for field_name in self.fields:
                alignment = meta_data.get(field_name, {}).get('alignment', 'left')
                verbose_name = meta_data.get(field_name, {}).get('verbose_name', '')
                row_data['fields'].append({
                    'content': self.table_field_content(instance, field_name),
                    'verbose_name': verbose_name,
                    'alignment': alignment,
                })
            data['body']['rows'].append(row_data)
        return data

    def table_meta_data(self, model) -> dict:
        return table_meta_data(model)

    def table_row_url(self, instance) -> str:
        return table_row_url(instance)

    def table_field_content(self, instance, field_name) -> str:
        return table_field_content(instance, field_name)

    def response(self, request, extra_content: str = None, replace_body: bool = False) -> HttpResponse:
        return super().response(request, extra_content, replace_body)


class TableRowResponse(Response):

    base_template = 'ui/table_row_update.html'

    def __init__(
        self, *,
        instance: models.Model,
        fields: list[str],
        instance_label: str | Promise | None = None,
        footer: dict = None,
        page: paginator.Page = None,
    ):
        super().__init__(template=self.base_template, context={})
        self.instance = instance
        self.instance.refresh_from_db()
        self.fields = fields
        self.instance_label = instance_label
        self.footer = footer or {}
        self.page = page

    def get_context(self):
        context = super().get_context()
        context.update({
            'instance': self.instance,
            'table_data': self.table_data(),
            'fields': self.fields,
            'instance_label': self.instance_label,
            'footer': self.footer,
            'page': self.page
        })
        return context

    def table_data(self):
        data = {
            'row': [],
            'footer': [],
        }
        data = {
            'header': {'fields': []},
            'body': {'rows': []},
            'footer': {'fields': []},
        }
        meta_data = self.table_meta_data(self.instance._meta.model)
        row_data = {
            'fields': [],
            'id': self.instance.pk,
            'is_active': None,
            'hx_get': None,
        }
        if self.instance_label:
            row_data['fields'].append({'content': str(self.instance), 'alignment': 'left'})
            data['footer']['fields'].append({'content': '', 'alignment': 'left'})
        for field_name in self.fields:
            alignment = meta_data.get(field_name, {}).get('alignment', 'left')
            verbose_name = meta_data.get(field_name, {}).get('verbose_name', '')
            row_data['fields'].append({
                'content': self.table_field_content(self.instance, field_name),
                'verbose_name': verbose_name,
                'alignment': alignment,
            })
            data['footer']['fields'].append({
                'content': self.footer.get(field_name, ''),
                'verbose_name': verbose_name,
                'alignment': alignment,
            })
        data['body']['rows'].append(row_data)
        return data

    def table_meta_data(self, model) -> dict:
        return table_meta_data(model)

    def table_field_content(self, instance, field_name) -> str:
        return table_field_content(instance, field_name)

    def response(self, request, extra_content: str = None, replace_body: bool = False) -> HttpResponse:
        return super().response(request, extra_content, replace_body)

    def render(self, request) -> str:
        res = super().render(request)
        if self.page:
            pagination_update = OobResponse(
                template='ui/layout.html#pagination',
                swap='innerHTML:#pagination',
                context=dict(page=self.page)
            ).render(request)
            res += pagination_update
        return res


def table_meta_data(model: models.Model):
    data = {}
    model_fields = model._meta.get_fields()
    number_fields = [
        'IntegerField',
        'FloatField',
        'DecimalField',
    ]
    for field in filter(lambda f: f.concrete, model_fields):
        data[field.name] = {
            'verbose_name': field.verbose_name,
            'alignment': field.get_internal_type() in number_fields and 'right' or 'left',
        }
    if hasattr(model, 'ui_table_data'):
        data.update(model.ui_table_data())
    return data


def table_row_url(instance) -> str:
    if not hasattr(instance, 'get_absolute_url'):
        return ''
    return instance.get_absolute_url()


def table_field_content(instance, field_name) -> str:
    field = getattr(instance, field_name)
    if isinstance(field, models.Manager):
        return mark_safe(f"""
            <span>
                <div class="tags">
                    {''.join(['<span class="tag">{}</span>'.format(o) for o in field.all()])}
                </div>
            </span>
        """)
    if getattr(field, 'choices', []):
        return getattr(instance, f'get_{field_name}_display')
    return field is not None and field or ''



import logging
import re
from datetime import datetime, date, timedelta

from django.contrib.auth import get_user_model
from django.shortcuts import resolve_url
from django.utils.functional import Promise
from django.utils.translation import gettext_lazy as _
from django.db.models import Manager
from django import template
from django.apps import apps
from django.template.loader import render_to_string
from django.utils.safestring import mark_safe
from django.forms import widgets
from accrete.contrib.ui.models import Theme
from accrete.tenant import get_tenant, unscoped
from accrete.models import Tenant, Member
from accrete.contrib.ui import config

_logger = logging.getLogger(__name__)
register = template.Library()
User = get_user_model()


@register.simple_tag(name='combine_templates')
def combine_templates(template_name, request=None):
    html = ''
    for app in apps.app_configs.keys():
        try:
            html += render_to_string(
                f'{app}/{template_name}',
                request=request
            )
        except template.TemplateDoesNotExist:
            continue
    return mark_safe(html)


@register.filter(name='related_obj_url')
def related_obj_url(param: object, value: str):
    attr = getattr(param, value, False)
    if attr and hasattr(attr, 'get_absolute_url'):
        return attr.get_absolute_url()
    return '#'


@register.filter(name='get_attr')
def get_attr_from_string(param: object, value: str):

    def return_save(attr):
        if isinstance(attr, Manager) and hasattr(attr, 'related_model'):
            return attr.related_model
        if attr is None:
            return None
        if isinstance(attr, str):
            return attr
        return attr

    try:
        attribute = getattr(param, value, False)
    except (AttributeError,):
        _logger.warning(f'Object {param} has no attribute {value}')
        attribute = None
    if attribute and callable(attribute):
        attribute = attribute()
    return return_save(attribute)


@register.filter(name='has_attr')
def has_attr(param: object, value: str):
    return hasattr(param, value)


@register.filter(name='get_item')
def get_item(param, value):
    return param[value]


@register.filter(name='get_model')
def get_model(instance):
    return instance._meta.model


@register.filter(name='message_class')
def message_class(message):
    if message.level == 25:
        return 'is-success'
    if message.level == 30:
        return 'is-warning'
    if message.level == 40:
        return 'is-danger'
    return ''


@register.filter(name='timedelta_cast')
def timedelta_cast(td: timedelta, code: str) -> str | None:
    if not isinstance(td, timedelta):
        return None
    codes = ['days', 'hours', 'minutes', 'seconds', 'microseconds']
    if code not in codes:
        return None
    return str(getattr(td, code))


@register.filter(name='weekday')
def datetime_to_weekday(dt: datetime | date | str, default=None) -> str | Promise:
    if dt is None:
        return default
    if isinstance(dt, str):
        dt = datetime.strptime(dt, '%Y-%m-%d')
    mapping = {
        1: _('Mon'),
        2: _('Tue'),
        3: _('Wed'),
        4: _('Thu'),
        5: _('Fri'),
        6: _('Sat'),
        7: _('Sun')
    }
    return mapping[dt.isoweekday()]


@register.filter(name='xrefsave')
def x_ref_save(param: str):
    return re.sub(r'[^A-Za-z]+', '', param)


@register.filter(name='wrap_form_field')
def wrap_form_field(field, icon=None):
    if isinstance(field.field.widget, widgets.Textarea):
        html = render_to_string(
            'ui/templatetags/field.html#textarea',
            {'field': field, 'icon': None}
        )
        return mark_safe(html)
    if isinstance(field.field.widget, widgets.CheckboxInput):
        html = render_to_string(
            'ui/templatetags/field.html#form_checkbox',
            {'field': field, 'icon': None}
        )
        return mark_safe(html)
    if isinstance(field.field.widget, widgets.RadioSelect):
        html = render_to_string(
            'ui/templatetags/field.html#form_radio',
            {'field': field, 'icon': None}
        )
        return mark_safe(html)
    html = render_to_string(
        'ui/templatetags/field.html#form_field',
        {'field': field, 'icon': icon}
    )
    return mark_safe(html)


@register.filter(name='wrap_model_field')
def wrap_model_field(instance, field_name):
    field = instance._meta.get_field(field_name)
    if field.choices:
        value = getattr(instance, f'get_{field_name}_display')()
    else:
        value = getattr(instance, field_name)
    html = render_to_string(
        'ui/templatetags/field.html#model_field', {
            'label': field.help_text or field.verbose_name,
            'value': value
        })
    return mark_safe(html)


@register.filter(name='default_if_falsy')
def default_if_falsy(value, default):
    if bool(value):
        return value
    return default


@register.simple_tag(name='custom_theme')
def custom_theme(user: User) -> str:
    if user.is_anonymous:
        return ''
    tenant = get_tenant()
    tenant_theme = Theme.objects.filter(
        tenant=tenant, tenant__isnull=False
    ).first()
    if tenant_theme and tenant_theme.force_tenant_theme:
        return mark_safe(tenant_theme.theme_markup)
    if user.theme == 'custom':
        theme = Theme.objects.filter(user=user).first()
        return theme and mark_safe(theme.theme_markup) or ''
    if user.theme == 'preset' and tenant_theme:
        return mark_safe(tenant_theme.theme_markup)
    return ''


@register.simple_tag(name='base_theme')
def base_theme(user: User) -> str:
    if user.is_anonymous:
        return 'light'
    tenant = get_tenant()
    tenant_theme = Theme.objects.filter(
        tenant=tenant, tenant__isnull=False
    ).first()
    if tenant_theme and tenant_theme.force_tenant_theme:
        return tenant_theme.base_theme
    if user.theme == 'custom':
        theme = Theme.objects.filter(user=user).first()
        return theme and theme.base_theme or 'light'
    if user.theme == 'preset' and tenant_theme:
        return tenant_theme.base_theme
    return user.theme


@register.simple_tag(name='theme_attr')
def theme_attribute(user: User, attribute: str) -> str:
    if user.is_anonymous:
        return None
    tenant = get_tenant()
    tenant_theme = Theme.objects.filter(
        tenant=tenant, tenant__isnull=False, force_tenant_theme=True,
    ).first()
    if tenant_theme:
        return getattr(tenant_theme, 'navbar_color', 'is-primary')
    if user.theme != 'custom':
        return 'is-primary'
    theme = user.theme == 'custom' and user.user_themes.first()
    color_class = theme and getattr(theme, 'navbar_color', None)
    if color_class is None:
        return ''
    return color_class


@register.simple_tag(name='tenant_quick_switch')
def tenant_quick_switch(user: User):
    templ = 'ui/templatetags/tenant_quick_switch.html'
    tenant = get_tenant()
    with unscoped():
        tenants = Tenant.objects.filter(
            members__in=Member.objects.filter(user=user)
        )
        if tenant:
            tenants = tenants.exclude(pk=tenant.pk)
    switch_url = getattr(config, 'ACCRETE_TENANT_QUICK_SWITCH_URL')
    if switch_url:
        switch_url = resolve_url(switch_url)
    return mark_safe(render_to_string(
        templ, context={
            'tenant': tenant,
            'tenants': tenants,
            'switch_url': switch_url
        }
    ))


@register.filter(name='filename')
def filename(param):
    return str(param).split('/')[-1]


@register.filter(name='zero_trail')
def zero_trail(value: str):
    if not isinstance(value, str):
        return value
    if not set(value).intersection(set(['.', ','])):
        return value
    value = value.rstrip('0') or ''
    if value and not value[-1].isdigit():
        value = value[:-1]
    return mark_safe(value)



# Generated by Django 4.2.6 on 2023-10-10 20:06

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        ('accrete', '0001_initial'),
    ]

    operations = [
        migrations.CreateModel(
            name='Sequence',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=50, verbose_name='Name')),
                ('nextval', models.PositiveBigIntegerField(default=1, verbose_name='Next Value')),
                ('step', models.PositiveIntegerField(default=1, verbose_name='Step')),
                ('tenant', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='accrete.tenant')),
            ],
            options={
                'db_table': 'accrete_sequence',
            },
        ),
        migrations.AddConstraint(
            model_name='sequence',
            constraint=models.UniqueConstraint(fields=('name', 'tenant'), name='unique_name_per_tenant'),
        ),
        migrations.AddConstraint(
            model_name='sequence',
            constraint=models.UniqueConstraint(fields=('name', 'tenant', 'nextval'), name='unique_val_per_name_tenant'),
        ),
    ]



from django.core.validators import MaxValueValidator
from django.db import models
from django.utils.translation import gettext_lazy as _

from accrete.models import TenantModel


class Sequence(TenantModel):

    class Meta:
        db_table = 'accrete_sequence'
        constraints = [
            models.UniqueConstraint(
                name='unique_name_per_tenant',
                fields=['name', 'tenant']
            ),
            models.UniqueConstraint(
                name='unique_val_per_name_tenant',
                fields=['name', 'tenant', 'nextval']
            )
        ]

    name = models.CharField(
        verbose_name=_('Name'),
        max_length=255,
    )

    nextval = models.PositiveBigIntegerField(
        verbose_name=_('Next Value'),
        default=1
    )

    step = models.PositiveIntegerField(
        verbose_name=_('Step'),
        default=1
    )

    prefix = models.CharField(
        verbose_name=_('Prefix'),
        default='',
    )

    padding = models.IntegerField(
        verbose_name=_('Padding'),
        validators=[MaxValueValidator(10)],
        default=0,
        help_text=_('Pads the sequence with zeros to the given length.')
    )

    has_subsequence_per_year = models.BooleanField(
        verbose_name=_('Subsequence per year'),
        default=False,
    )

    def __str__(self):
        return self.name



import json
from django.test import TestCase
from django.contrib.auth import get_user_model
from accrete.contrib import ui
from accrete.models import Tenant, Member

User = get_user_model()


class FilterTestCase(TestCase):

    def setUp(self):
        super().setUp()
        self.tenant1 = Tenant(name='tenant 1')
        self.tenant1.save()
        self.tenant2 = Tenant(name='tenant2')
        self.tenant2.save()

        self.pezi = User(email='pezi@test.com', username='pezi')
        self.pezi.save()
        self.dagobert = User(email='dagobert@test.com', username='dagobert')
        self.dagobert.save()

        self.member_pezi = Member(tenant=self.tenant1, user=self.pezi)
        self.member_pezi.save()
        self.member_dagobert = Member(tenant=self.tenant1, user=self.dagobert)
        self.member_dagobert.save()
        self.member_dagobert2 = Member(tenant=self.tenant2, user=self.dagobert)
        self.member_dagobert2.save()

    def test_get_page(self):
        f = ui.Filter(
            model=Member,
            query_dict={'q': json.dumps({'tenant__id__exact': self.tenant1.pk})},
        )
        self.assertQuerySetEqual(
            f.get_page().object_list,
            Member.objects.filter(
                pk__in=[self.member_pezi.pk, self.member_dagobert.pk]
            )
        )
        f = ui.Filter(
            model=Member,
            query_dict={'q': json.dumps({'tenant__id__exact': self.tenant2.pk})},
        )
        self.assertQuerySetEqual(
            f.get_page().object_list,
            Member.objects.filter(pk__in=[self.member_dagobert2.pk])
        )



# Generated by Django 5.2.7 on 2025-10-04 06:20

import json
import accrete.fields
from django.db import migrations
from django.conf import settings


def char_to_translated_char(apps, schema):
    AccessGroup = apps.get_model('accrete', 'AccessGroup')
    default_lang = getattr(settings, 'LANGUAGE_CODE', 'en-us')
    for group in AccessGroup.objects.all():
        try:
            description = json.dumps(json.loads(group.description).get(default_lang))
            if not description:
                description = None
        except Exception:
            description = group.description or None
        group.description = description
        group.save()


def translated_to_char(apps, schema):
    AccessGroup = apps.get_model('accrete', 'AccessGroup')
    default_lang = getattr(settings, 'LANGUAGE_CODE', 'en-us')
    for group in AccessGroup.objects.all():
        try:
            group.description = json.loads(group.description).get(default_lang, '')
        except Exception:
            continue


class Migration(migrations.Migration):

    dependencies = [
        ('accrete', '0009_alter_accessgroup_name'),
    ]

    operations = [
        migrations.RunPython(char_to_translated_char, translated_to_char),
        migrations.AlterField(
            model_name='accessgroup',
            name='description',
            field=accrete.fields.TranslatedTextField(blank=True, null=True, verbose_name='Description'),
        ),
    ]



# Generated by Django 5.2.1 on 2025-08-08 06:27

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('ui', '0003_alter_theme_check_user_or_tenant'),
    ]

    operations = [
        migrations.AddField(
            model_name='theme',
            name='force_tenant_theme',
            field=models.BooleanField(default=False, help_text='Apply this theme regardless of user preferences', verbose_name='Force Tenant Theme'),
        ),
    ]



from django.http.response import HttpResponseRedirectBase
from django.shortcuts import HttpResponse
from django.test import TestCase
from django.contrib.auth import get_user_model
from django.test.client import RequestFactory
from django.conf import settings

from accrete.decorator import tenant_required
from accrete.models import Tenant, Member

User = get_user_model()


def group_missing_action(request, tenant_groups, member_groups):
    return HttpResponse('group missing action')


@tenant_required()
def dummy(request):
    return True


@tenant_required(member_groups=['test'], group_missing_action=group_missing_action)
def dummy_group(request):
    return True


class DecoratorTestCase(TestCase):

    def setUp(self):
        self.factory = RequestFactory()
        self.user = User(email='user@accrete.test', username='Test User', is_active=True)
        self.user.save()
        self.tenant = Tenant(name='Test Tenant')
        self.tenant.save()

    def test_tenant_not_set(self):
        request = self.factory.get('/')
        request.user = self.user
        request.tenant = None
        response = dummy(request)
        self.assertTrue(isinstance(response, HttpResponseRedirectBase))
        self.assertTrue(response.url.startswith(settings.ACCRETE_TENANT_NOT_SET_URL))

    def test_group_missing(self):
        member = Member(tenant=self.tenant, user=self.user)
        member.save()
        request = self.factory.get('/')
        request.user = self.user
        request.tenant = self.tenant
        request.member = member
        result = dummy_group(request)
        self.assertEqual(result.text, 'group missing action')

    def test_staff_bypass(self):
        user = User(
            email='staff@test.com',
            username='Staff User',
            is_active=True,
            is_staff=True
        )
        user.save()
        request = self.factory.get('/')
        request.tenant = self.tenant
        request.member = None
        request.user = user
        result = dummy_group(request)
        self.assertEqual(result, True)



# Generated by Django 5.2.1 on 2025-08-07 17:11

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('ui', '0001_initial'),
    ]

    operations = [
        migrations.AlterField(
            model_name='theme',
            name='color_danger',
            field=models.CharField(default='#FF6584', max_length=7, verbose_name='Danger Color'),
        ),
        migrations.AlterField(
            model_name='theme',
            name='color_link',
            field=models.CharField(default='#65D0FE', max_length=7, verbose_name='Link Color'),
        ),
        migrations.AlterField(
            model_name='theme',
            name='color_primary',
            field=models.CharField(default='#46C68D', max_length=7, verbose_name='Primary Color'),
        ),
        migrations.AlterField(
            model_name='theme',
            name='color_success',
            field=models.CharField(default='#46C68D', max_length=7, verbose_name='Success Color'),
        ),
        migrations.AlterField(
            model_name='theme',
            name='color_warning',
            field=models.CharField(default='#FEB600', max_length=7, verbose_name='Warning Color'),
        ),
    ]



# Generated by Django 5.2.8 on 2026-02-07 13:13

import django.core.validators
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('sequence', '0003_alter_sequence_tenant'),
    ]

    operations = [
        migrations.AddField(
            model_name='sequence',
            name='has_subsequence_per_year',
            field=models.BooleanField(default=False, verbose_name='Subsequence per year'),
        ),
        migrations.AddField(
            model_name='sequence',
            name='padding',
            field=models.IntegerField(default=0, help_text='Pads the sequence with zeros to the given length.', validators=[django.core.validators.MaxLengthValidator(10)], verbose_name='Padding'),
        ),
        migrations.AddField(
            model_name='sequence',
            name='prefix',
            field=models.CharField(default='', verbose_name='Prefix'),
        ),
    ]



from django.test import TestCase
from accrete.contrib.ui.templatetags import ui as tags


class TemplatetagsTestCase(TestCase):

    def test_zero_trail_no_str(self):
        value = None
        res = tags.zero_trail(value)
        self.assertIsNone(res)
        value = 50
        res = tags.zero_trail(value)
        self.assertEqual(value, res)

    def test_zero_trail(self):
        values = [
            ('', ''), ('1', '1'), ('10', '10'), ('01', '01'),
            ('010', '010'), ('1.10', '1.1'), ('1,11', '1,11'),
            ('1.01', '1.01'), ('1.010', '1.01'), ('1.0', '1')
        ]
        for value in values:
            res = tags.zero_trail(value[0])
            self.assertEqual(res, value[1])




from django.core import paginator
from django.db import models
from django.http import HttpResponse
from django.utils.translation import gettext_lazy as _
from django.utils.functional import Promise
from accrete.contrib.ui import Filter
from .response import Response, OobResponse
from .window import WindowResponse, WindowResponseConfig


class ListResponse(WindowResponse):

    def __init__(
        self, *,
        title: str | Promise,
        context: dict,
        list_entry_template: str = None,
        page: paginator.Page = None,
        ui_filter: Filter = None,
        endless_scroll: bool = False,
        header_template: str = None,
        panel_template: str = None,
        column_count: int = 1,
        column_height: str | None = '150px',
        overview_template: str = 'ui/list.html',
        detail_header_template: str = None,
        detail_data_template: str = None,
        detail_enabled: bool = True,
        config: WindowResponseConfig = None
    ):
        assert page is not None or ui_filter is not None, _(
            'Argument page or ui_filter must be supplied'
        )
        super().__init__(
            title=title,
            context=context,
            overview_template=overview_template,
            header_template=header_template,
            panel_template=panel_template,
            config=config
        )
        self.page = page or (ui_filter and ui_filter.get_page())
        self.list_entry_template = list_entry_template
        self.ui_filter = ui_filter
        self.endless_scroll = endless_scroll
        self.column_count = column_count
        self.column_height = column_height
        self.detail_header_template = detail_header_template
        self.detail_data_template = detail_data_template
        self.detail_enabled = detail_enabled

    def _has_panel(self):
        return bool(self.panel_template or self.ui_filter)

    def get_context(self):
        context = super().get_context()
        context.update({
            'list_entry_template': self.list_entry_template,
            'page': self.page,
            'ui_filter': self.ui_filter,
            'endless_scroll': self.endless_scroll,
            'column_count': self.column_count,
            'column_height': self.column_height,
            'detail_enabled': self.detail_enabled,
            'detail_header_template': self.detail_header_template,
            'detail_data_template': self.detail_data_template,
            'show_content_right': str(bool(
                self.detail_header_template or self.detail_data_template
            )).lower(),
        })
        return context

    def response(self, request, extra_content: str = None, replace_body: bool = False) -> HttpResponse:
        return super().response(request, extra_content, replace_body)


class ListEntryResponse(Response):

    base_template = 'ui/list_update.html'

    def __init__(
        self, *,
        instance: models.Model,
        list_entry_template: str,
        context: dict = None,
        page: paginator.Page = None,
        is_new: bool = False,
        column_count: int = 1,
        column_height: str = '150px',
    ):
        super().__init__(template=self.base_template, context=context or {})
        self.instance = instance
        self.instance.refresh_from_db()
        self.list_entry_template = list_entry_template
        self.page = page
        self.is_new = is_new
        self.column_count = column_count
        self.column_height = column_height

    def get_context(self):
        context = super().get_context()
        context.update({
            'instance': self.instance,
            'list_entry_template': self.list_entry_template,
            'is_new': self.is_new,
            'column_count': self.column_count,
            'column_height': self.column_height
        })
        return context

    def render(self, request) -> str:
        res = super().render(request)
        if self.page:
            pagination_update = OobResponse(
                template='ui/layout.html#pagination',
                swap='innerHTML:#pagination',
                context=dict(page=self.page)
            ).render(request)
            res += pagination_update
        return res



import os

from django.http import HttpResponse, HttpResponseNotFound
from django.shortcuts import get_object_or_404
from django.conf import settings
from accrete.models import Tenant, Member
from accrete.decorator import tenant_required


@tenant_required()
def get_tenant_file(request, tenant_id, filepath):
    tenant = get_object_or_404(Tenant, pk=tenant_id)
    if not request.user.is_staff:
        member = Member.objects.filter(user=request.user, tenant=tenant)
        if not member.exists():
            return HttpResponseNotFound()
    filepath = f'{settings.MEDIA_ROOT}/{tenant_id}/{filepath}'
    if not os.path.exists(filepath):
        return HttpResponseNotFound()
    with open(filepath, 'rb') as f:
        return HttpResponse(f)



from django.urls import path

from .views import filter

app_name = 'ui'

urlpatterns = [
    path('filter/<str:model>/params/', filter.params, name='filter_params'),
    path('filter/<str:model>/input/', filter.set_filter_input, name='filter_input'),
    path('filter/<str:model>/query/add/', filter.filter_add_query, name='filter_add_query')
]



from . import widgets
from .filter import Filter
from .views.decorator import tenant_required
from .responses.detail import DetailResponse
from .responses.list import (
    ListResponse,
    ListEntryResponse,
)
from .responses.modal import (
    ModalResponse,
    ModalMessageResponse,
)
from .responses.response import (
    Response,
    OobResponse,
    TriggerResponse,
    ClientTrigger,
    update,
    add_trigger,
    search_select_response,
    message_response,
    redirect_response
)
from .responses.table import (
    TableResponse,
    TableRowResponse,
)
from .responses.window import (
    WindowResponse,
    WindowResponseConfig,
)




import re
from django.utils.functional import Promise
from .response import Response


class ModalResponse(Response):

    def __init__(
        self, *,
        modal_id: str,
        template: str,
        context: dict,
        title: str | Promise = None,
        is_update: bool = False,
        is_blocking: bool = False,
        modal_width: str = None

    ):
        super().__init__(template=template, context=context)
        self.modal_id = modal_id
        self.title = title
        self.is_update = is_update
        self.is_blocking = is_blocking
        self.modal_width = modal_width

    def get_context(self):
        context = super().get_context()
        context.update({
            'title': self.title,
            'modal_id': re.sub(r'[^A-Za-z-]+', '', self.modal_id).strip('-'),
            'is_update': self.is_update,
            'is_blocking': self.is_blocking,
            'modal_width': self.modal_width
        })
        return context


class ModalMessageResponse(ModalResponse):

    def __init__(
        self, *,
        modal_id: str,
        template: str = None,
        message: str | Promise = None,
        title: str | Promise = None,
        modal_width: str = None,
    ):
        super().__init__(
            modal_id=modal_id,
            template=template or 'ui/modal_message.html',
            context={'message': message},
            title=title,
            modal_width=modal_width,
        )



from django.core.exceptions import ValidationError
from django.db import models
from django.conf import settings
from django.utils.translation import gettext_lazy as _
from accrete.tenant import get_tenant
from accrete.managers import TenantManager, MemberManager, AccessGroupManager
from accrete.fields import TranslatedCharField, TranslatedTextField


class TenantModel(models.Model):

    class Meta:
        abstract = True

    tenant = models.ForeignKey(
        verbose_name=_('Tenant'),
        to='accrete.Tenant',
        on_delete=models.CASCADE
    )

    objects = TenantManager()

    def save(
        self,
        *args,
        force_insert=False,
        force_update=False,
        using=None,
        update_fields=None
    ):
        tenant = get_tenant()
        if self.pk and tenant and self.tenant_id != tenant.pk:
            raise ValueError('Current tenant differs from tenant of the record!')
        if self.pk and not self.tenant and not tenant:
            raise ValueError(
                'Tenant not provided! '
                'Use accrete.tenant.set_tenant() '
                'or set the tenant explicitly on the instance.'
            )
        if tenant:
            self.tenant_id = tenant.pk
        super().save(
            force_insert=force_insert,
            force_update=force_update,
            using=using,
            update_fields=update_fields
        )

    # used in accrete.contrib.ui added here only for convenience
    @staticmethod
    def exclude_from_filter():
        return ['tenant']


class Tenant(models.Model):

    class Meta:
        verbose_name = _('Tenant')
        verbose_name_plural = _('Tenants')
        ordering = ['-is_active', 'name']
        db_table = 'accrete_tenant'

    name = models.CharField(
        verbose_name=_('Name'),
        max_length=255
    )

    is_active = models.BooleanField(
        verbose_name=_('Active'),
        default=True
    )

    access_groups = models.ManyToManyField(
        to='accrete.AccessGroup',
        through='accrete.TenantAccessGroupRel',
        through_fields=('tenant', 'access_group'),
        blank=True
    )

    def __str__(self):
        return self.name


class Member(models.Model):

    class Meta:
        verbose_name = _('Member')
        verbose_name_plural = _('Members')
        ordering = ['tenant', 'user']
        db_table = 'accrete_member'

    user = models.ForeignKey(
        to=settings.AUTH_USER_MODEL,
        related_name='memberships',
        on_delete=models.CASCADE
    )

    tenant = models.ForeignKey(
        to='accrete.Tenant',
        related_name='members',
        on_delete=models.CASCADE
    )

    is_active = models.BooleanField(
        verbose_name=_('Active'),
        default=True
    )

    access_groups = models.ManyToManyField(
        to='accrete.AccessGroup',
        through='accrete.MemberAccessGroupRel',
        through_fields=('member', 'access_group'),
        blank=True
    )

    objects = MemberManager()

    def __str__(self):
        return f'{self.user}'


class AccessGroup(models.Model):

    class Meta:
        verbose_name = _('Access Group')
        verbose_name_plural = _('Access Groups')
        ordering = ['name']
        db_table = 'accrete_access_group'
        constraints = [
            models.UniqueConstraint(
                name='unique_code',
                fields=['code']
            )
        ]

    objects = AccessGroupManager()

    name = TranslatedCharField(
        verbose_name=_('Name')
    )

    description = TranslatedTextField(
        verbose_name=_('Description'),
        null=True,
        blank=True
    )

    code = models.CharField(
        verbose_name=_('Code'),
        max_length=100
    )

    apply_on = models.CharField(
        verbose_name=_('Apply On'),
        max_length=10,
        choices=[
            ('tenant', _('Tenant')),
            ('member', _('Member'))
        ],
        default='tenant'
    )

    def __str__(self):
        return f'{self.name} ({self.get_apply_on_display()})'


class MemberAccessGroupRel(models.Model):

    class Meta:
        verbose_name = _('Member Access Group Relation')
        verbose_name_plural = _('Member Access Group Relations')
        ordering = ['member']
        db_table = 'accrete_member_access_group_rel'
        constraints = [
            models.UniqueConstraint(
                name='unique_member_per_group',
                fields=['member', 'access_group']
            )
        ]

    member = models.ForeignKey(
        to='accrete.Member',
        on_delete=models.CASCADE
    )

    access_group = models.ForeignKey(
        to='accrete.AccessGroup',
        on_delete=models.CASCADE
    )

    def __str__(self):
        return f'{self.member} - {self.access_group}'

    def clean(self):
        if self.access_group.apply_on != 'member':
            raise ValidationError(_('Access Group must apply on members'))


class TenantAccessGroupRel(models.Model):

    class Meta:
        verbose_name = _('Tenant Access Group Relation')
        verbose_name_plural = _('Tenant Access Groups Relations')
        ordering = ['tenant']
        db_table = 'accrete_tenant_access_group_rel'
        constraints = [
            models.UniqueConstraint(
                name='unique_tenant_per_group',
                fields=['tenant', 'access_group']
            )
        ]

    tenant = models.ForeignKey(
        to='accrete.Tenant',
        on_delete=models.CASCADE
    )

    access_group = models.ForeignKey(
        to='accrete.AccessGroup',
        on_delete=models.CASCADE
    )

    def __str__(self):
        return f'{self.tenant} - {self.access_group}'

    def clean(self):
        if self.access_group.apply_on != 'tenant':
            raise ValidationError(_('Access Group must apply on tenants'))



from datetime import date

from django.db.models import Q
from freezegun import freeze_time
from django.conf import settings
from django.test import SimpleTestCase, TestCase
from django.utils import timezone, translation
from accrete.forms import TenantModelForm
from accrete.utils import dates, forms, models, views
from accrete.models import AccessGroup, Member, Tenant


class DatesSimpleTestCase(SimpleTestCase):

    @freeze_time('2026-03-03')
    def test_current_year_start(self):
        self.assertEqual(
            dates.current_year_start(),
            timezone.now().replace(
                year=2026, month=1, day=1, hour=0, second=0, microsecond=0
            )
        )

    @freeze_time('2026-03-03')
    def test_current_year_end(self):
        self.assertEqual(
            dates.current_year_end(),
            timezone.now().replace(
                year=2026,
                month=12,
                day=31,
                hour=23,
                minute=59,
                second=59,
                microsecond=999999,
            )
        )

    @freeze_time('2026-03-03')
    def test_current_month_start(self):
        self.assertEqual(
            dates.current_month_start(),
            timezone.now().replace(
                year=2026, month=3, day=1, hour=0, second=0, microsecond=0
            )
        )

    @freeze_time('2026-02-03')
    def test_current_month_end(self):
        self.assertEqual(
            dates.current_month_end(),
            timezone.now().replace(
                year=2026,
                month=2,
                day=28,
                hour=23,
                minute=59,
                second=59,
                microsecond=999999,
            )
        )

    @freeze_time('2026-03-03', tz_offset=0)
    def test_next_relative_month_start(self):
        self.assertEqual(
            dates.next_relative_month_start(date.today()),
            date(year=2026, month=4, day=1)
        )

    @freeze_time('2026-03-03')
    def test_next_relative_year_start(self):
        self.assertEqual(
            dates.next_relative_year_start(date.today()),
            date(year=2027, month=1, day=1)
        )

    @freeze_time('2026-03-03')
    def test_day_of_month_or_last(self):
        self.assertEqual(
            dates.day_of_month_or_last(2026, 3, 3),
            3
        )

    @freeze_time('2026-02-28')
    def test_day_of_month_or_last(self):
        self.assertEqual(
            dates.day_of_month_or_last(2026, 2, 29),
            28
        )

    @freeze_time('2024-02-29')
    def test_day_of_month_or_last_leap(self):
        self.assertEqual(
            dates.day_of_month_or_last(2024, 2, 29),
            29
        )


class FormsTestCase(TestCase):

    class TestModelForm(TenantModelForm):

        class Meta:
            model = AccessGroup
            fields = [
                'name',
                'description',
                'code',
                'apply_on'
            ]

    def test_save_form_valid(self):
        data = {
            'name': 'Test',
            'description': 'Test Group',
            'code': 'test',
            'apply_on': 'tenant',
        }
        result = forms.save_form(self.TestModelForm(data=data))
        self.assertEqual(True, result.is_saved)

    def test_save_form_invalid(self):
        data = {
            'description': 'Test Group',
            'code': 'test',
            'apply_on': 'tenant',
        }
        result = forms.save_form(self.TestModelForm(data=data))
        self.assertEqual(False, result.is_saved)


class ModelTestCase(TestCase):

    def test_get_related_model(self):
        translation.activate('en')
        result = models.get_related_model(Member, 'tenant__access_groups')
        self.assertEqual(result, (AccessGroup, ['Member', 'Tenant', 'Access Group']))

    def test_get_related_field(self):
        result = models.get_related_field(Member, 'tenant__access_groups__name')
        self.assertEqual(result, AccessGroup._meta.get_field('name'))
    
    def test_translated_db_value(self):
        translation.activate(settings.LANGUAGE_CODE)
        group = AccessGroup(name='Test', code='test')
        group.save()
        result = models.translated_db_value(instance=group, field='name')
        self.assertDictEqual(result, {settings.LANGUAGE_CODE: 'Test'})

    def test_model_to_dict(self):
        translation.activate(settings.LANGUAGE_CODE)
        group = AccessGroup(name='Test', description='Test Group',  code='test')
        group.save()
        result = models.model_to_dict(
            instance=group, 
            fields=['name', 'description', 'code'], 
            translated=False,
        )
        self.assertDictEqual(result, {
            'name': {settings.LANGUAGE_CODE: 'Test'},
            'description': {settings.LANGUAGE_CODE: 'Test Group'},
            'code': 'test',
        })

    def test_model_to_dict_translated(self):
        translation.activate(settings.LANGUAGE_CODE)
        group = AccessGroup(name='Test', description='Test Group',  code='test')
        group.save()
        result = models.model_to_dict(
            instance=group, 
            fields=['name', 'description', 'code'], 
            translated=True,
        
        )
        self.assertDictEqual(result, {
            'name': 'Test',
            'description': 'Test Group',
            'code': 'test',
        })


class ViewTestCase(TestCase):
    
    def test_page_from_querystring(self):
        group = AccessGroup(name='Test', description='Test Group',  code='test')
        group.save()
        result = views.page_from_querystring(
            AccessGroup, query_dict={'q': '[{"code": "test"}]'}
        )
        self.assertQuerySetEqual(
            result.object_list, AccessGroup.objects.filter(code='test')
        )

    def test_parse_querystring(self):
        query = '[{"is_active": true}, "|", {"tenant__name__iexact": "Test Tenant"}]'
        self.assertEqual(
            views.parse_querystring(Member, query),
            Q(is_active=True) | Q(tenant__name__iexact='Test Tenant')
        )
        query = '[{"is_active": true},  "|", [{"is_active": false}, {"tenant__name__startswith": "Test"}]]'
        self.assertEqual(
            views.parse_querystring(Member, query),
            Q(is_active=True) | Q(is_active=False, tenant__name__startswith='Test')
        )



from django.db import models
from django.conf import settings


class UserConfirmationToken(models.Model):
    class Meta:
        db_table = 'user_registration_token'

    user = models.OneToOneField(
        to=settings.AUTH_USER_MODEL,
        on_delete=models.CASCADE,
        related_name='token'
    )

    token = models.CharField(
        verbose_name='Token',
        max_length=255
    )



import json
import logging
import datetime
from decimal import Decimal

from django.db.models import QuerySet, Q
from django.http.request import QueryDict
from django.db.models.fields import Field
from django.template.loader import render_to_string
from django.utils.translation import gettext_lazy as _
from django.utils.safestring import mark_safe
from django.apps import apps
from django.core import paginator
from accrete.utils.models import get_related_model, get_related_field
from accrete.utils import page_from_querystring, filter_from_querystring

_logger = logging.getLogger(__name__)


class Filter:

    LABEL_EXACT = _('Equals')
    LABEL_EXACT_NOT = _('Equals Not')
    LABEL_ICONTAINS = _('Contains')
    LABEL_ICONTAINS_NOT = _('Contains Not')
    LABEL_STARTS_WITH = _('Starts With')
    LABEL_ENDS_WITH = _('Ends With')
    LABEL_GTE = _('Greater or Equal')
    LABEL_LTE = _('Less or Equal')
    LABEL_TRUE = _('True')
    LABEL_FALSE = _('False')
    LABEL_SET = _('Is Set')
    LABEL_NOT_SET = _('Is Not Set')
    LABEL_AND = _('And')
    LABEL_OR = _('Or')
    LABEL_XOR = _('Not Or')

    DATE_FORMAT = '%Y-%m-%d'

    TYPES_INTEGER = [
        'AutoField',
        'BigAutoField',
        'IntegerField',
        'PositiveSmallIntegerField',
    ]
    TYPES_NUMBER = ['DecimalField', 'FloatField']
    TYPES_CHAR = ['CharField', 'TextField', 'JSONField']
    TYPES_BOOL = ['BooleanField']
    TYPES_DATETIME = ['DateTimeField']
    TYPES_DATE = ['DateField']
    TYPES_TIME = ['TimeField']

    def __init__(
        self,
        model,
        query_dict: QueryDict,
        default_lookup: str = None,
        query: Q = None,
        select_related: list[str] = None,
        prefetch_related: list[str] = None
    ):
        self.model = model
        self.query_dict = query_dict
        self.model_name = f'{self.model._meta.app_label}.{self.model._meta.model_name}'
        self.query = query or Q()
        self.select_related = select_related
        self.prefetch_related = prefetch_related
        seen_models = query_dict.get('models', '').split(',')
        self.seen_models = [
            model for model in seen_models
            if model and model != self.model_name
        ] + [self.model_name]
        self.default_lookup = default_lookup
        self.exclude = []
        if hasattr(self.model, 'exclude_from_filter'):
            self.exclude.extend(self.model.exclude_from_filter())

    def get_page(
        self,
        select_related: list = None,
        prefetch_related: list = None,
        query: Q = None
    ) -> paginator.Page:
        return page_from_querystring(
            self.model,
            self.query_dict,
            select_related=select_related or self.select_related,
            prefetch_related=prefetch_related or self.prefetch_related,
            query=query or self.query
        )

    def get_queryset(self) -> QuerySet:
        return filter_from_querystring(self.model, self.query_dict).filter(self.query)

    def query_params(self):
        fields = filter(
            lambda x: x.name not in self.exclude,
            self.model._meta.get_fields()
        )
        params = []
        path = self.query_dict.get('path', '')
        for field in fields:
            field_params = self._get_field_params(field)
            if not field.is_relation or field_params['model_name'] not in self.seen_models:
                params.append(self._get_field_params(field))
        has_previous = len(self.seen_models) > 1
        ctx = dict(
            params=sorted(params, key=lambda x: x['label'].lower()),
            model_name=self.model_name,
            seen_models=','.join(self.seen_models),
            verbose_model_name=self.model._meta.verbose_name,
            has_previous=has_previous,
        )
        if has_previous:
            previous_model_name = self.seen_models[-2]
            previous_path = '__'.join(path.split('__')[:-1])
            ctx.update(
                previous_model_name=previous_model_name,
                previous_seen_models=','.join(
                    self.seen_models[:-1] if has_previous else []
                ),
                previous_verbose_model_name=apps.get_model(
                    *previous_model_name.split('.')
                )._meta.verbose_name,
                previous_path=previous_path
            )
        return render_to_string('ui/filter/query_params.html', ctx)

    def query_input(self, lookup: str = None):
        lookup = lookup or self.default_lookup
        if not lookup:
            ctx = {
                'verbose_lookup': str(_('Select an attribute')),
                'query_dict': self.query_dict,
                'model_name': self.model_name
            }
            return render_to_string('ui/filter/query_input.html', ctx)
        prefix = lookup.startswith('~') and '~' or ''
        lookup = lookup.removeprefix('~')
        lookup_parts = lookup.split('__')
        lookup_operator = lookup_parts[-1]
        rel_path = '__'.join(lookup_parts[:-2])
        model, names = get_related_model(self.model, rel_path)
        field = model._meta.get_field(lookup_parts[-2])
        input_params = self._get_input_params(field, lookup_operator)
        ctx = {
            'verbose_lookup': ' > '.join(self._get_query_tag_lhs(prefix + lookup)),
            'field': field,
            'lookup': prefix + lookup,
            'input': input_params,
            'lookup_operator': lookup_operator,
            'query_dict': self.query_dict,
            'model_name': self.model_name
        }
        return render_to_string('ui/filter/query_input.html', ctx)

    def query_tags(self, query: list | dict = None, operator: str = None):
        if not query:
            query = json.loads(self.query_dict.get('q', '[]'))
        if isinstance(query, dict):
            query = [query]
        operator = operator or '&'
        html = '<div class="query-group-container">'
        html += render_to_string(
            'ui/filter/query_operator.html', {'operator': operator}
        )
        html += (
            '<div class="query-group" '
            'x-sort="applyQuery();" x-sort:group="query" '
            'x-sort:config="{preventOnFilter: false}">'
        )
        for idx, item in enumerate(query):
            if isinstance(item, str):
                continue
            operator = '&'
            if idx > 0 and isinstance(query[idx - 1], str):
                operator = query[idx - 1]
            if isinstance(item, list):
                html += self.query_tags(item, operator)
            if isinstance(item, dict):
                item_ctx = self._get_query_tag_context(item, operator)
                html += render_to_string('ui/filter/query_tags.html', item_ctx)
        html += '</div></div>'
        return mark_safe(html)

    def _get_input_params(self, field: Field, lookup_operator: str):
        field_type = field.get_internal_type()
        params = {
            'field_type': field_type,
            'data_type': self._internal_type_to_data_type(field_type, lookup_operator)
        }
        if lookup_operator == 'isnull':
            params.update(
                input_type='select',
                choices=[('false', str(_('True'))), ('true', str(_('False')))],
                data_type='bool'
            )
            return params
        if field_type in self.TYPES_BOOL:
            params.update(
                input_type='select',
                choices=[('true', str(_('True'))), ('false', str(_('False')))],
                data_type='bool'
            )
            return params
        elif field_type in self.TYPES_INTEGER:
            if field.choices:
                params.update(input_type='select', choices=field.choices)
            else:
                params.update(input_type='number', step=1)
        elif field_type in self.TYPES_NUMBER:
            step = (
                hasattr(field, 'decimal_places')
                and self._cast_decimal_places_to_step(field.decimal_places)
                or 1
            )
            params.update(input_type='number', step=step)
        elif field_type in self.TYPES_CHAR:
            if field.choices:
                params.update(input_type='select', choices=field.choices)
            else:
                params.update(input_type='text')
        elif field_type in self.TYPES_DATETIME:
            params.update(input_type='datetime-local', format=self.DATE_FORMAT)
        elif field_type in self.TYPES_DATE:
            params.update(input_type='date', format=self.DATE_FORMAT)
        elif field_type in self.TYPES_TIME:
            params.update(input_type='time')
        return params

    def _internal_type_to_data_type(self, internal_type, lookup: str):
        if lookup == 'isnull' or internal_type in self.TYPES_BOOL:
            return 'bool'
        if internal_type in self.TYPES_INTEGER + self.TYPES_NUMBER:
            return 'number'
        return 'text'

    def _get_query_tag_context(self, data: dict, operator: str):
        ctx = {'tags': []}
        first_operator = operator
        operator = '&'
        for key, value in data.items():

            field_path = key.split('__')
            field = get_related_field(
                self.model, '__'.join(field_path[:-1]).removeprefix('~')
            )
            lookup = field_path[-1]
            internal_type = field.get_internal_type()
            data_type = self._internal_type_to_data_type(internal_type, lookup)
            ctx['tags'].append({
                'lhs': self._get_query_tag_lhs(key),
                'rhs': self._get_query_tag_rhs(value, internal_type, lookup),
                'lookup': key,
                'value': value,
                'operator': operator,
                'data_type': data_type,
                'model_name': self.model_name
            })
        if ctx['tags']:
            ctx['tags'][0]['operator'] = first_operator
        return ctx

    def _get_query_tag_lhs(self, lookup: str):
        prefix = lookup.startswith('~') and '~' or ''
        lookup = lookup.removeprefix('~')
        parts = lookup.split('__')
        assert len(parts) >= 2
        related_parts = parts[:-2]
        rel_model, names = get_related_model(
            self.model, '__'.join(related_parts)
        )
        try:
            names.extend([
                str(rel_model._meta.get_field(parts[-2]).verbose_name),
                str(self._get_lookup_label(f'{prefix}{parts[-1]}'))
            ])
        except AttributeError:
            names.extend([
                str(rel_model._meta.get_field(parts[-2]).field.verbose_name),
                str(self._get_lookup_label(f'{prefix}{parts[-1]}'))
            ])
        return names[1:]

    def _get_query_tag_rhs(self, value, internal_type, lookup):
        if lookup == 'isnull':
            return not value
        if internal_type in self.TYPES_DATE:
            return datetime.date.fromisoformat(value)
        if internal_type in self.TYPES_DATETIME:
            return datetime.datetime.fromisoformat(value)
        if internal_type in self.TYPES_NUMBER:
            return Decimal(value)
        return value

    def _get_lookup_label(self, lookup: str):
        labels = {
            'exact': self.LABEL_EXACT,
            '~exact': self.LABEL_EXACT_NOT,
            'icontains': self.LABEL_ICONTAINS,
            '~icontains': self.LABEL_ICONTAINS_NOT,
            'istartswith': self.LABEL_STARTS_WITH,
            'iendswith': self.LABEL_ENDS_WITH,
            'gte': self.LABEL_GTE,
            'lte': self.LABEL_LTE,
            'isnull': self.LABEL_SET,
            '~isnull': self.LABEL_NOT_SET
        }
        return str(labels.get(lookup, lookup))

    def _get_field_params(self, field):
        path = self.query_dict.get('path', '')
        label = ''
        help_text = ''
        name = field.name
        if path:
            field_path = f'{path}__{name}'
        else:
            field_path = name
        field_type = field.get_internal_type()
        if field.concrete or not field.is_relation:
            label = str(field.verbose_name)
            help_text = field.help_text
        if field.is_relation:
            choices = None
            step = None
            model = ''
            if field_type in ['ForeignKey', 'OneToOneField']:
                model = (
                    f'{field.related_model._meta.app_label}.'
                    f'{field.related_model._meta.model_name}'
                )
                if not field.concrete:
                    label = (
                        f'{field.related_model._meta.verbose_name_plural} / '
                        f'{field.field.verbose_name}'
                    )
                else:
                    help_text = field.help_text
            elif field_type == 'ManyToManyField':
                model = (
                    f'{field.remote_field.model._meta.app_label}.'
                    f'{field.remote_field.model._meta.model_name}'
                )
                if not field.concrete:
                    label = (
                        f'{field.remote_field.model._meta.verbose_name} / '
                        f'{field.remote_field.verbose_name}'
                    )
        else:
            step = (
                hasattr(field, 'decimal_places')
                and self._cast_decimal_places_to_step(field.decimal_places)
                or 1
            )
            choices = field.choices or None
            model = self.model_name
        return {
            'field': field,
            'name': name,
            'field_path': field_path,
            'label': label,
            'type': field_type,
            'param_type': '',
            'choices': choices,
            'null': field.null,
            'step': step,
            'is_relation': field.is_relation,
            'model_name': model,
            'help_text': help_text
        }

    @staticmethod
    def _cast_decimal_places_to_step(decimal_places):
        if not decimal_places or decimal_places < 1:
            return '1'
        zero_count = decimal_places - 1
        return f'0.{"0" * zero_count}1'



from .response import Response, add_trigger


class DetailResponse(Response):

    base_template = 'ui/detail.html'

    def __init__(
        self, *,
        context: dict,
        header_template: str = None,
        data_template: str = None
    ):
        super().__init__(template=self.base_template, context=context)
        self.header_template = header_template
        self.data_template = data_template
        self.context.update()

    def add_trigger(self, response):
        super().add_trigger(response)
        add_trigger(response, 'activate-detail')

    def get_context(self):
        context = super().get_context()
        context.update({
            'detail_header_template': self.header_template,
            'detail_data_template': self.data_template
        })
        return context



# Generated by Django 5.1.4 on 2025-02-19 14:32
import logging
from django.db import migrations
from django.db import connection

_logger = logging.getLogger(__name__)


def create_countries(apps, schema_editor):
    with connection.cursor() as cr:
        cr.execute(
            """
            INSERT INTO 
                accrete_country(name, iso_code_a2, iso_code_a3, vat_prefix, order_priority)
                VALUES 
                    ('{"de-de": "Aruba", "en-us": "Aruba"}', 'AW', 'ABW', null, 0),
                    ('{"de-de": "Afghanistan", "en-us": "Afghanistan"}', 'AF', 'AFG', null, 0),
                    ('{"de-de": "Angola", "en-us": "Angola"}', 'AO', 'AGO', null, 0),
                    ('{"de-de": "Anguilla", "en-us": "Anguilla"}', 'AI', 'AIA', null, 0),
                    ('{"de-de": "Åland Islands", "en-us": "Åland Islands"}', 'AX', 'ALA', null, 0),
                    ('{"de-de": "Albanien", "en-us": "Albania"}', 'AL', 'ALB', null, 0),
                    ('{"de-de": "Andorra", "en-us": "Andorra"}', 'AD', 'AND', null, 0),
                    ('{"de-de": "Vereinigte Arabische Emirate", "en-us": "United Arab Emirates"}', 'AE', 'ARE', null, 0),
                    ('{"de-de": "Argentinien", "en-us": "Argentina"}', 'AR', 'ARG', null, 0),
                    ('{"de-de": "Armenien", "en-us": "Armenia"}', 'AM', 'ARM', null, 0),
                    ('{"de-de": "American Samoa", "en-us": "American Samoa"}', 'AS', 'ASM', null, 0),
                    ('{"de-de": "Antarctica", "en-us": "Antarctica"}', 'AQ', 'ATA', null, 0),
                    ('{"de-de": "French Southern Territories", "en-us": "French Southern Territories"}', 'TF', 'ATF', null, 0),
                    ('{"de-de": "Antigua and Barbuda", "en-us": "Antigua and Barbuda"}', 'AG', 'ATG', null, 0),
                    ('{"de-de": "Australien", "en-us": "Australia"}', 'AU', 'AUS', null, 0),
                    ('{"de-de": "Österreich", "en-us": "Austria"}', 'AT', 'AUT', 'AT', 0),
                    ('{"de-de": "Aserbaidschan", "en-us": "Azerbaijan"}', 'AZ', 'AZE', null, 0),
                    ('{"de-de": "Burundi", "en-us": "Burundi"}', 'BI', 'BDI', null, 0),
                    ('{"de-de": "Belgien", "en-us": "Belgium"}', 'BE', 'BEL', 'BE', 0),
                    ('{"de-de": "Benin", "en-us": "Benin"}', 'BJ', 'BEN', null, 0),
                    ('{"de-de": "Bonaire, Sint Eustatius and Saba", "en-us": "Bonaire, Sint Eustatius and Saba"}', 'BQ', 'BES', null, 0),
                    ('{"de-de": "Burkina Faso", "en-us": "Burkina Faso"}', 'BF', 'BFA', null, 0),
                    ('{"de-de": "Bangladesch", "en-us": "Bangladesh"}', 'BD', 'BGD', null, 0),
                    ('{"de-de": "Bulgarien", "en-us": "Bulgaria"}', 'BG', 'BGR', 'BG', 0),
                    ('{"de-de": "Bahrain", "en-us": "Bahrain"}', 'BH', 'BHR', null, 0),
                    ('{"de-de": "Bahamas", "en-us": "Bahamas"}', 'BS', 'BHS', null, 0),
                    ('{"de-de": "Bosnien und Herzegowina", "en-us": "Bosnia and Herzegovina"}', 'BA', 'BIH', null, 0),
                    ('{"de-de": "Saint Barthélemy", "en-us": "Saint Barthélemy"}', 'BL', 'BLM', null, 0),
                    ('{"de-de": "Belarus", "en-us": "Belarus"}', 'BY', 'BLR', null, 0),
                    ('{"de-de": "Belize", "en-us": "Belize"}', 'BZ', 'BLZ', null, 0),
                    ('{"de-de": "Bermuda", "en-us": "Bermuda"}', 'BM', 'BMU', null, 0),
                    ('{"de-de": "Bolivien", "en-us": "Bolivia, Plurinational State of"}', 'BO', 'BOL', null, 0),
                    ('{"de-de": "Brasilien", "en-us": "Brazil"}', 'BR', 'BRA', null, 0),
                    ('{"de-de": "Barbados", "en-us": "Barbados"}', 'BB', 'BRB', null, 0),
                    ('{"de-de": "Brunei Darussalam", "en-us": "Brunei Darussalam"}', 'BN', 'BRN', null, 0),
                    ('{"de-de": "Bhutan", "en-us": "Bhutan"}', 'BT', 'BTN', null, 0),
                    ('{"de-de": "Bouvet Island", "en-us": "Bouvet Island"}', 'BV', 'BVT', null, 0),
                    ('{"de-de": "Botswana", "en-us": "Botswana"}', 'BW', 'BWA', null, 0),
                    ('{"de-de": "Central African Republic", "en-us": "Central African Republic"}', 'CF', 'CAF', null, 0),
                    ('{"de-de": "Kanada", "en-us": "Canada"}', 'CA', 'CAN', null, 0),
                    ('{"de-de": "Cocos (Keeling) Islands", "en-us": "Cocos (Keeling) Islands"}', 'CC', 'CCK', null, 0),
                    ('{"de-de": "Schweiz", "en-us": "Switzerland"}', 'CH', 'CHE', null, 0),
                    ('{"de-de": "Chile", "en-us": "Chile"}', 'CL', 'CHL', null, 0),
                    ('{"de-de": "China", "en-us": "China"}', 'CN', 'CHN', null, 0),
                    ('{"de-de": "Elfenbeinküste", "en-us": "Ivory Coast"}', 'CI', 'CIV', null, 0),
                    ('{"de-de": "Cameroon", "en-us": "Cameroon"}', 'CM', 'CMR', null, 0),
                    ('{"de-de": "Congo, The Democratic Republic of the", "en-us": "Congo, The Democratic Republic of the"}', 'CD', 'COD', null, 0),
                    ('{"de-de": "Congo", "en-us": "Congo"}', 'CG', 'COG', null, 0),
                    ('{"de-de": "Cook Islands", "en-us": "Cook Islands"}', 'CK', 'COK', null, 0),
                    ('{"de-de": "Kolumbien", "en-us": "Colombia"}', 'CO', 'COL', null, 0),
                    ('{"de-de": "Comoros", "en-us": "Comoros"}', 'KM', 'COM', null, 0),
                    ('{"de-de": "Cabo Verde", "en-us": "Cabo Verde"}', 'CV', 'CPV', null, 0),
                    ('{"de-de": "Costa Rica", "en-us": "Costa Rica"}', 'CR', 'CRI', null, 0),
                    ('{"de-de": "Kuba", "en-us": "Cuba"}', 'CU', 'CUB', null, 0),
                    ('{"de-de": "Curaçao", "en-us": "Curaçao"}', 'CW', 'CUW', null, 0),
                    ('{"de-de": "Christmas Island", "en-us": "Christmas Island"}', 'CX', 'CXR', null, 0),
                    ('{"de-de": "Cayman Islands", "en-us": "Cayman Islands"}', 'KY', 'CYM', null, 0),
                    ('{"de-de": "Zypern", "en-us": "Cyprus"}', 'CY', 'CYP', 'CY', 0),
                    ('{"de-de": "Tschechien", "en-us": "Czechia"}', 'CZ', 'CZE', 'CZ', 0),
                    ('{"de-de": "Deutschland", "en-us": "Germany"}', 'DE', 'DEU', 'DE', 0),
                    ('{"de-de": "Djibouti", "en-us": "Djibouti"}', 'DJ', 'DJI', null, 0),
                    ('{"de-de": "Dominica", "en-us": "Dominica"}', 'DM', 'DMA', null, 0),
                    ('{"de-de": "Dänemark", "en-us": "Denmark"}', 'DK', 'DNK', 'DK', 0),
                    ('{"de-de": "Dominican Republic", "en-us": "Dominican Republic"}', 'DO', 'DOM', null, 0),
                    ('{"de-de": "Algerien", "en-us": "Algeria"}', 'DZ', 'DZA', null, 0),
                    ('{"de-de": "Ecuador", "en-us": "Ecuador"}', 'EC', 'ECU', null, 0),
                    ('{"de-de": "Ägypten", "en-us": "Egypt"}', 'EG', 'EGY', null, 0),
                    ('{"de-de": "Eritrea", "en-us": "Eritrea"}', 'ER', 'ERI', null, 0),
                    ('{"de-de": "Western Sahara", "en-us": "Western Sahara"}', 'EH', 'ESH', null, 0),
                    ('{"de-de": "Spanien", "en-us": "Spain"}', 'ES', 'ESP', 'ES', 0),
                    ('{"de-de": "Estland", "en-us": "Estonia"}', 'EE', 'EST', 'EE', 0),
                    ('{"de-de": "Ethiopia", "en-us": "Ethiopia"}', 'ET', 'ETH', null, 0),
                    ('{"de-de": "Finnland", "en-us": "Finland"}', 'FI', 'FIN', 'FI', 0),
                    ('{"de-de": "Fiji", "en-us": "Fiji"}', 'FJ', 'FJI', null, 0),
                    ('{"de-de": "Falkland Islands (Malvinas)", "en-us": "Falkland Islands (Malvinas)"}', 'FK', 'FLK', null, 0),
                    ('{"de-de": "Frankreich", "en-us": "France"}', 'FR', 'FRA', 'FR', 0),
                    ('{"de-de": "Faroe Islands", "en-us": "Faroe Islands"}', 'FO', 'FRO', null, 0),
                    ('{"de-de": "Micronesia, Federated States of", "en-us": "Micronesia, Federated States of"}', 'FM', 'FSM', null, 0),
                    ('{"de-de": "Gabon", "en-us": "Gabon"}', 'GA', 'GAB', null, 0),
                    ('{"de-de": "Vereinigtes Königreich", "en-us": "United Kingdom"}', 'GB', 'GBR', null, 0),
                    ('{"de-de": "Georgia", "en-us": "Georgia"}', 'GE', 'GEO', null, 0),
                    ('{"de-de": "Guernsey", "en-us": "Guernsey"}', 'GG', 'GGY', null, 0),
                    ('{"de-de": "Ghana", "en-us": "Ghana"}', 'GH', 'GHA', null, 0),
                    ('{"de-de": "Gibraltar", "en-us": "Gibraltar"}', 'GI', 'GIB', null, 0),
                    ('{"de-de": "Guinea", "en-us": "Guinea"}', 'GN', 'GIN', null, 0),
                    ('{"de-de": "Guadeloupe", "en-us": "Guadeloupe"}', 'GP', 'GLP', null, 0),
                    ('{"de-de": "Gambia", "en-us": "Gambia"}', 'GM', 'GMB', null, 0),
                    ('{"de-de": "Guinea-Bissau", "en-us": "Guinea-Bissau"}', 'GW', 'GNB', null, 0),
                    ('{"de-de": "Equatorial Guinea", "en-us": "Equatorial Guinea"}', 'GQ', 'GNQ', null, 0),
                    ('{"de-de": "Griechenland", "en-us": "Greece"}', 'GR', 'GRC', null, 0),
                    ('{"de-de": "Grenada", "en-us": "Grenada"}', 'GD', 'GRD', null, 0),
                    ('{"de-de": "Greenland", "en-us": "Greenland"}', 'GL', 'GRL', null, 0),
                    ('{"de-de": "Guatemala", "en-us": "Guatemala"}', 'GT', 'GTM', null, 0),
                    ('{"de-de": "French Guiana", "en-us": "French Guiana"}', 'GF', 'GUF', null, 0),
                    ('{"de-de": "Guam", "en-us": "Guam"}', 'GU', 'GUM', null, 0),
                    ('{"de-de": "Guyana", "en-us": "Guyana"}', 'GY', 'GUY', null, 0),
                    ('{"de-de": "Hong Kong", "en-us": "Hong Kong"}', 'HK', 'HKG', null, 0),
                    ('{"de-de": "Heard Island and McDonald Islands", "en-us": "Heard Island and McDonald Islands"}', 'HM', 'HMD', null, 0),
                    ('{"de-de": "Honduras", "en-us": "Honduras"}', 'HN', 'HND', null, 0),
                    ('{"de-de": "Kroatien", "en-us": "Croatia"}', 'HR', 'HRV', 'HR', 0),
                    ('{"de-de": "Haiti", "en-us": "Haiti"}', 'HT', 'HTI', null, 0),
                    ('{"de-de": "Ungarn", "en-us": "Hungary"}', 'HU', 'HUN', 'HU', 0),
                    ('{"de-de": "Indonesien", "en-us": "Indonesia"}', 'ID', 'IDN', null, 0),
                    ('{"de-de": "Isle of Man", "en-us": "Isle of Man"}', 'IM', 'IMN', null, 0),
                    ('{"de-de": "Indien", "en-us": "India"}', 'IN', 'IND', null, 0),
                    ('{"de-de": "British Indian Ocean Territory", "en-us": "British Indian Ocean Territory"}', 'IO', 'IOT', null, 0),
                    ('{"de-de": "Irland", "en-us": "Ireland"}', 'IE', 'IRL', 'IE', 0),
                    ('{"de-de": "Iran, Islamic Republic of", "en-us": "Iran, Islamic Republic of"}', 'IR', 'IRN', null, 0),
                    ('{"de-de": "Iraq", "en-us": "Iraq"}', 'IQ', 'IRQ', null, 0),
                    ('{"de-de": "Iceland", "en-us": "Iceland"}', 'IS', 'ISL', null, 0),
                    ('{"de-de": "Israel", "en-us": "Israel"}', 'IL', 'ISR', null, 0),
                    ('{"de-de": "Italien", "en-us": "Italy"}', 'IT', 'ITA', 'IT', 0),
                    ('{"de-de": "Jamaica", "en-us": "Jamaica"}', 'JM', 'JAM', null, 0),
                    ('{"de-de": "Jersey", "en-us": "Jersey"}', 'JE', 'JEY', null, 0),
                    ('{"de-de": "Jordan", "en-us": "Jordan"}', 'JO', 'JOR', null, 0),
                    ('{"de-de": "Japan", "en-us": "Japan"}', 'JP', 'JPN', null, 0),
                    ('{"de-de": "Kasachstan", "en-us": "Kazakhstan"}', 'KZ', 'KAZ', null, 0),
                    ('{"de-de": "Kenia", "en-us": "Kenya"}', 'KE', 'KEN', null, 0),
                    ('{"de-de": "Kyrgyzstan", "en-us": "Kyrgyzstan"}', 'KG', 'KGZ', null, 0),
                    ('{"de-de": "Cambodia", "en-us": "Cambodia"}', 'KH', 'KHM', null, 0),
                    ('{"de-de": "Kiribati", "en-us": "Kiribati"}', 'KI', 'KIR', null, 0),
                    ('{"de-de": "Saint Kitts and Nevis", "en-us": "Saint Kitts and Nevis"}', 'KN', 'KNA', null, 0),
                    ('{"de-de": "Südkorea", "en-us": "Korea, Republic of"}', 'KR', 'KOR', null, 0),
                    ('{"de-de": "Kuwait", "en-us": "Kuwait"}', 'KW', 'KWT', null, 0),
                    ('{"de-de": "Lao Peoples Democratic Republic", "en-us": "Lao Peoples Democratic Republic"}', 'LA', 'LAO', null, 0),
                    ('{"de-de": "Lebanon", "en-us": "Lebanon"}', 'LB', 'LBN', null, 0),
                    ('{"de-de": "Liberia", "en-us": "Liberia"}', 'LR', 'LBR', null, 0),
                    ('{"de-de": "Libya", "en-us": "Libya"}', 'LY', 'LBY', null, 0),
                    ('{"de-de": "Saint Lucia", "en-us": "Saint Lucia"}', 'LC', 'LCA', null, 0),
                    ('{"de-de": "Liechtenstein", "en-us": "Liechtenstein"}', 'LI', 'LIE', null, 0),
                    ('{"de-de": "Sri Lanka", "en-us": "Sri Lanka"}', 'LK', 'LKA', null, 0),
                    ('{"de-de": "Lesotho", "en-us": "Lesotho"}', 'LS', 'LSO', null, 0),
                    ('{"de-de": "Litauen", "en-us": "Lithuania"}', 'LT', 'LTU', 'LT', 0),
                    ('{"de-de": "Luxemburg", "en-us": "Luxembourg"}', 'LU', 'LUX', 'LU', 0),
                    ('{"de-de": "Lettland", "en-us": "Latvia"}', 'LV', 'LVA', 'LV', 0),
                    ('{"de-de": "Macao", "en-us": "Macao"}', 'MO', 'MAC', null, 0),
                    ('{"de-de": "Saint Martin (French part)", "en-us": "Saint Martin (French part)"}', 'MF', 'MAF', null, 0),
                    ('{"de-de": "Marokko", "en-us": "Morocco"}', 'MA', 'MAR', null, 0),
                    ('{"de-de": "Monaco", "en-us": "Monaco"}', 'MC', 'MCO', null, 0),
                    ('{"de-de": "Moldova, Republic of", "en-us": "Moldova, Republic of"}', 'MD', 'MDA', null, 0),
                    ('{"de-de": "Madagascar", "en-us": "Madagascar"}', 'MG', 'MDG', null, 0),
                    ('{"de-de": "Maldives", "en-us": "Maldives"}', 'MV', 'MDV', null, 0),
                    ('{"de-de": "Mexiko", "en-us": "Mexico"}', 'MX', 'MEX', null, 0),
                    ('{"de-de": "Marshall Islands", "en-us": "Marshall Islands"}', 'MH', 'MHL', null, 0),
                    ('{"de-de": "North Macedonia", "en-us": "North Macedonia"}', 'MK', 'MKD', null, 0),
                    ('{"de-de": "Mali", "en-us": "Mali"}', 'ML', 'MLI', null, 0),
                    ('{"de-de": "Malta", "en-us": "Malta"}', 'MT', 'MLT', 'MT', 0),
                    ('{"de-de": "Myanmar", "en-us": "Myanmar"}', 'MM', 'MMR', null, 0),
                    ('{"de-de": "Montenegro", "en-us": "Montenegro"}', 'ME', 'MNE', null, 0),
                    ('{"de-de": "Mongolia", "en-us": "Mongolia"}', 'MN', 'MNG', null, 0),
                    ('{"de-de": "Northern Mariana Islands", "en-us": "Northern Mariana Islands"}', 'MP', 'MNP', null, 0),
                    ('{"de-de": "Mozambique", "en-us": "Mozambique"}', 'MZ', 'MOZ', null, 0),
                    ('{"de-de": "Mauritania", "en-us": "Mauritania"}', 'MR', 'MRT', null, 0),
                    ('{"de-de": "Montserrat", "en-us": "Montserrat"}', 'MS', 'MSR', null, 0),
                    ('{"de-de": "Martinique", "en-us": "Martinique"}', 'MQ', 'MTQ', null, 0),
                    ('{"de-de": "Mauritius", "en-us": "Mauritius"}', 'MU', 'MUS', null, 0),
                    ('{"de-de": "Malawi", "en-us": "Malawi"}', 'MW', 'MWI', null, 0),
                    ('{"de-de": "Malaysia", "en-us": "Malaysia"}', 'MY', 'MYS', null, 0),
                    ('{"de-de": "Mayotte", "en-us": "Mayotte"}', 'YT', 'MYT', null, 0),
                    ('{"de-de": "Namibia", "en-us": "Namibia"}', 'NA', 'NAM', null, 0),
                    ('{"de-de": "New Caledonia", "en-us": "New Caledonia"}', 'NC', 'NCL', null, 0),
                    ('{"de-de": "Niger", "en-us": "Niger"}', 'NE', 'NER', null, 0),
                    ('{"de-de": "Norfolk Island", "en-us": "Norfolk Island"}', 'NF', 'NFK', null, 0),
                    ('{"de-de": "Nigeria", "en-us": "Nigeria"}', 'NG', 'NGA', null, 0),
                    ('{"de-de": "Nicaragua", "en-us": "Nicaragua"}', 'NI', 'NIC', null, 0),
                    ('{"de-de": "Niue", "en-us": "Niue"}', 'NU', 'NIU', null, 0),
                    ('{"de-de": "Niederlande", "en-us": "Netherlands"}', 'NL', 'NLD', 'NL', 0),
                    ('{"de-de": "Norwegen", "en-us": "Norway"}', 'NO', 'NOR', null, 0),
                    ('{"de-de": "Nepal", "en-us": "Nepal"}', 'NP', 'NPL', null, 0),
                    ('{"de-de": "Nauru", "en-us": "Nauru"}', 'NR', 'NRU', null, 0),
                    ('{"de-de": "Neuseeland", "en-us": "New Zealand"}', 'NZ', 'NZL', null, 0),
                    ('{"de-de": "Oman", "en-us": "Oman"}', 'OM', 'OMN', null, 0),
                    ('{"de-de": "Pakistan", "en-us": "Pakistan"}', 'PK', 'PAK', null, 0),
                    ('{"de-de": "Panama", "en-us": "Panama"}', 'PA', 'PAN', null, 0),
                    ('{"de-de": "Pitcairn", "en-us": "Pitcairn"}', 'PN', 'PCN', null, 0),
                    ('{"de-de": "Peru", "en-us": "Peru"}', 'PE', 'PER', null, 0),
                    ('{"de-de": "Philippinen", "en-us": "Philippines"}', 'PH', 'PHL', null, 0),
                    ('{"de-de": "Palau", "en-us": "Palau"}', 'PW', 'PLW', null, 0),
                    ('{"de-de": "Papua New Guinea", "en-us": "Papua New Guinea"}', 'PG', 'PNG', null, 0),
                    ('{"de-de": "Polen", "en-us": "Poland"}', 'PL', 'POL', 'PL', 0),
                    ('{"de-de": "Puerto Rico", "en-us": "Puerto Rico"}', 'PR', 'PRI', null, 0),
                    ('{"de-de": "Nordkorea", "en-us": "North Korea"}', 'KP', 'PRK', null, 0),
                    ('{"de-de": "Portugal", "en-us": "Portugal"}', 'PT', 'PRT', 'PT', 0),
                    ('{"de-de": "Paraguay", "en-us": "Paraguay"}', 'PY', 'PRY', null, 0),
                    ('{"de-de": "Palestine, State of", "en-us": "Palestine, State of"}', 'PS', 'PSE', null, 0),
                    ('{"de-de": "French Polynesia", "en-us": "French Polynesia"}', 'PF', 'PYF', null, 0),
                    ('{"de-de": "Qatar", "en-us": "Qatar"}', 'QA', 'QAT', null, 0),
                    ('{"de-de": "Réunion", "en-us": "Réunion"}', 'RE', 'REU', null, 0),
                    ('{"de-de": "Rumänien", "en-us": "Romania"}', 'RO', 'ROU', 'RO', 0),
                    ('{"de-de": "Russland", "en-us": "Russian Federation"}', 'RU', 'RUS', null, 0),
                    ('{"de-de": "Rwanda", "en-us": "Rwanda"}', 'RW', 'RWA', null, 0),
                    ('{"de-de": "Saudi-Arabien", "en-us": "Saudi Arabia"}', 'SA', 'SAU', null, 0),
                    ('{"de-de": "Sudan", "en-us": "Sudan"}', 'SD', 'SDN', null, 0),
                    ('{"de-de": "Senegal", "en-us": "Senegal"}', 'SN', 'SEN', null, 0),
                    ('{"de-de": "Singapur", "en-us": "Singapore"}', 'SG', 'SGP', null, 0),
                    ('{"de-de": "South Georgia and the South Sandwich Islands", "en-us": "South Georgia and the South Sandwich Islands"}', 'GS', 'SGS', null, 0),
                    ('{"de-de": "Saint Helena, Ascension and Tristan da Cunha", "en-us": "Saint Helena, Ascension and Tristan da Cunha"}', 'SH', 'SHN', null, 0),
                    ('{"de-de": "Svalbard and Jan Mayen", "en-us": "Svalbard and Jan Mayen"}', 'SJ', 'SJM', null, 0),
                    ('{"de-de": "Solomon Islands", "en-us": "Solomon Islands"}', 'SB', 'SLB', null, 0),
                    ('{"de-de": "Sierra Leone", "en-us": "Sierra Leone"}', 'SL', 'SLE', null, 0),
                    ('{"de-de": "El Salvador", "en-us": "El Salvador"}', 'SV', 'SLV', null, 0),
                    ('{"de-de": "San Marino", "en-us": "San Marino"}', 'SM', 'SMR', null, 0),
                    ('{"de-de": "Somalia", "en-us": "Somalia"}', 'SO', 'SOM', null, 0),
                    ('{"de-de": "Saint Pierre and Miquelon", "en-us": "Saint Pierre and Miquelon"}', 'PM', 'SPM', null, 0),
                    ('{"de-de": "Serbien", "en-us": "Serbia"}', 'RS', 'SRB', null, 0),
                    ('{"de-de": "South Sudan", "en-us": "South Sudan"}', 'SS', 'SSD', null, 0),
                    ('{"de-de": "Sao Tome and Principe", "en-us": "Sao Tome and Principe"}', 'ST', 'STP', null, 0),
                    ('{"de-de": "Suriname", "en-us": "Suriname"}', 'SR', 'SUR', null, 0),
                    ('{"de-de": "Slowakei", "en-us": "Slovakia"}', 'SK', 'SVK', 'SK', 0),
                    ('{"de-de": "Slowenien", "en-us": "Slovenia"}', 'SI', 'SVN', 'SI', 0),
                    ('{"de-de": "Schweden", "en-us": "Sweden"}', 'SE', 'SWE', 'SE', 0),
                    ('{"de-de": "Eswatini", "en-us": "Eswatini"}', 'SZ', 'SWZ', null, 0),
                    ('{"de-de": "Sint Maarten (Dutch part)", "en-us": "Sint Maarten (Dutch part)"}', 'SX', 'SXM', null, 0),
                    ('{"de-de": "Seychelles", "en-us": "Seychelles"}', 'SC', 'SYC', null, 0),
                    ('{"de-de": "Syrian Arab Republic", "en-us": "Syrian Arab Republic"}', 'SY', 'SYR', null, 0),
                    ('{"de-de": "Turks and Caicos Islands", "en-us": "Turks and Caicos Islands"}', 'TC', 'TCA', null, 0),
                    ('{"de-de": "Chad", "en-us": "Chad"}', 'TD', 'TCD', null, 0),
                    ('{"de-de": "Togo", "en-us": "Togo"}', 'TG', 'TGO', null, 0),
                    ('{"de-de": "Thailand", "en-us": "Thailand"}', 'TH', 'THA', null, 0),
                    ('{"de-de": "Tajikistan", "en-us": "Tajikistan"}', 'TJ', 'TJK', null, 0),
                    ('{"de-de": "Tokelau", "en-us": "Tokelau"}', 'TK', 'TKL', null, 0),
                    ('{"de-de": "Turkmenistan", "en-us": "Turkmenistan"}', 'TM', 'TKM', null, 0),
                    ('{"de-de": "Timor-Leste", "en-us": "Timor-Leste"}', 'TL', 'TLS', null, 0),
                    ('{"de-de": "Tonga", "en-us": "Tonga"}', 'TO', 'TON', null, 0),
                    ('{"de-de": "Trinidad and Tobago", "en-us": "Trinidad and Tobago"}', 'TT', 'TTO', null, 0),
                    ('{"de-de": "Tunisia", "en-us": "Tunisia"}', 'TN', 'TUN', null, 0),
                    ('{"de-de": "Türkei", "en-us": "Turkey"}', 'TR', 'TUR', null, 0),
                    ('{"de-de": "Tuvalu", "en-us": "Tuvalu"}', 'TV', 'TUV', null, 0),
                    ('{"de-de": "Taiwan, Province of China", "en-us": "Taiwan, Province of China"}', 'TW', 'TWN', null, 0),
                    ('{"de-de": "Tanzania, United Republic of", "en-us": "Tanzania, United Republic of"}', 'TZ', 'TZA', null, 0),
                    ('{"de-de": "Uganda", "en-us": "Uganda"}', 'UG', 'UGA', null, 0),
                    ('{"de-de": "Ukraine", "en-us": "Ukraine"}', 'UA', 'UKR', null, 0),
                    ('{"de-de": "United States Minor Outlying Islands", "en-us": "United States Minor Outlying Islands"}', 'UM', 'UMI', null, 0),
                    ('{"de-de": "Uruguay", "en-us": "Uruguay"}', 'UY', 'URY', null, 0),
                    ('{"de-de": "Vereinigte Staaten", "en-us": "United States"}', 'US', 'USA', null, 0),
                    ('{"de-de": "Uzbekistan", "en-us": "Uzbekistan"}', 'UZ', 'UZB', null, 0),
                    ('{"de-de": "Holy See (Vatican City State)", "en-us": "Holy See (Vatican City State)"}', 'VA', 'VAT', null, 0),
                    ('{"de-de": "Saint Vincent and the Grenadines", "en-us": "Saint Vincent and the Grenadines"}', 'VC', 'VCT', null, 0),
                    ('{"de-de": "Venezuela, Bolivarian Republic of", "en-us": "Venezuela, Bolivarian Republic of"}', 'VE', 'VEN', null, 0),
                    ('{"de-de": "Virgin Islands, British", "en-us": "Virgin Islands, British"}', 'VG', 'VGB', null, 0),
                    ('{"de-de": "Virgin Islands, U.S.", "en-us": "Virgin Islands, U.S."}', 'VI', 'VIR', null, 0),
                    ('{"de-de": "Vietnam", "en-us": "Viet Nam"}', 'VN', 'VNM', null, 0),
                    ('{"de-de": "Vanuatu", "en-us": "Vanuatu"}', 'VU', 'VUT', null, 0),
                    ('{"de-de": "Wallis and Futuna", "en-us": "Wallis and Futuna"}', 'WF', 'WLF', null, 0),
                    ('{"de-de": "Samoa", "en-us": "Samoa"}', 'WS', 'WSM', null, 0),
                    ('{"de-de": "Yemen", "en-us": "Yemen"}', 'YE', 'YEM', null, 0),
                    ('{"de-de": "Südafrika", "en-us": "South Africa"}', 'ZA', 'ZAF', null, 0),
                    ('{"de-de": "Zambia", "en-us": "Zambia"}', 'ZM', 'ZMB', null, 0),
                    ('{"de-de": "Zimbabwe", "en-us": "Zimbabwe"}', 'ZW', 'ZWE', null, 0)
            """
        )


def revert(apps, schema_editor):
    Country = apps.get_model('country', 'Country')
    countries = Country.objects.all()
    try:
        countries.delete()
    except Exception as e:
        _logger.exception(e)


class Migration(migrations.Migration):

    dependencies = [
        ('country', '0001_initial'),
    ]

    operations = [
        migrations.RunPython(create_countries, revert)
    ]



# Generated by Django 5.2.1 on 2025-05-24 09:10

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('accrete', '0007_accessgroup_description'),
    ]

    operations = [
        migrations.AlterField(
            model_name='member',
            name='access_groups',
            field=models.ManyToManyField(blank=True, through='accrete.MemberAccessGroupRel', through_fields=('member', 'access_group'), to='accrete.accessgroup'),
        ),
        migrations.AlterField(
            model_name='tenant',
            name='access_groups',
            field=models.ManyToManyField(blank=True, through='accrete.TenantAccessGroupRel', through_fields=('tenant', 'access_group'), to='accrete.accessgroup'),
        ),
    ]



from django.test import TestCase
from accrete.models import Tenant
from accrete.tenant import set_tenant
from accrete.contrib.sequence.models import Sequence
from accrete.contrib.sequence import get_nextval


class SequenceTestCase(TestCase):

    def setUp(self):
        super().setUp()
        tenant = Tenant(name='test')
        tenant.save()
        set_tenant(tenant)

    def test_sequence_new(self):
        sequence_data = get_nextval('test_new', create_values={})
        self.assertEqual(sequence_data.counter, 1)
    
    def test_sequence_per_year_new(self):
        sequence_data = get_nextval(
            'test_per_year_new', 
            create_values={'has_subsequence_per_year': True}
        )
        self.assertEqual(sequence_data.counter, 1)
        base_sequence = Sequence.objects.get(name='test_per_year_new')
        self.assertEqual(base_sequence.nextval, 1)
        sequence_for_year = Sequence.objects.get(name='test_per_year_new__2026')
        self.assertEqual(sequence_for_year.nextval, 2)
        sequence_data = get_nextval('test_per_year_new')
        self.assertEqual(sequence_data.counter, 2)



import logging
from urllib.parse import urlparse, parse_qs
from django.contrib.auth import get_user_model

from accrete.models import Tenant, Member
from accrete.tenant import set_member, set_tenant

_logger = logging.getLogger(__name__)


class TenantMiddleware:
    """
    Set tenant and member contextvar
    """

    def __init__(self, inner):
        self.inner = inner

    async def __call__(self, scope, receive, send):
        user = scope.get('user')
        if not user or not isinstance(user, get_user_model()):
            set_member(None)
            return await self.inner(scope, receive, send)
        tenant_id = parse_qs(urlparse(
            f"?{scope.get('query_string', b'').decode()}"
        ).query).get('tenant_id', [None])[0]
        tenant = tenant_id and await Tenant.objects.aget(pk=tenant_id)
        if tenant and user.is_staff:
            set_tenant(tenant)
            return await self.inner(scope, receive, send)
        get_kwargs = {'user': user, 'tenant': tenant}
        try:
            member = await Member.objects.aget(**get_kwargs)
        except (Member.DoesNotExist, Member.MultipleObjectsReturned) as e:
            _logger.error(e)
            member = None
        set_member(member)
        return await self.inner(scope, receive, send)



# Generated by Django 5.1.4 on 2025-02-19 16:59

import accrete.fields
import django.db.models.functions.text
from django.db import migrations, models


class Migration(migrations.Migration):

    initial = True

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='Country',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', accrete.fields.TranslatedCharField(verbose_name='Name')),
                ('iso_code_a2', models.CharField(max_length=2, verbose_name='ISO Code Alpha 2')),
                ('iso_code_a3', models.CharField(max_length=3, verbose_name='ISO Code Alpha 3')),
                ('vat_prefix', models.CharField(blank=True, max_length=3, null=True, verbose_name='VAT ID Prefix')),
                ('order_priority', models.IntegerField(default=0, verbose_name='Order Priority')),
            ],
            options={
                'verbose_name': 'Country',
                'verbose_name_plural': 'Countries',
                'db_table': 'accrete_country',
                'ordering': ['-order_priority', django.db.models.functions.text.Lower('name')],
            },
        ),
    ]



from django.utils.translation import gettext_lazy as _
from django.apps import apps
from accrete.contrib import ui


def is_htmx(request):
    return request.headers.get('HX-Request', 'false') == 'true'


def get_group_data(tenant_groups, member_groups) -> dict:

    def _flat_groups() -> tuple[list[str], list[str]]:
        def group_list(g):
            if isinstance(g, str):
                return [g]
            elif isinstance(g, tuple):
                return [x for x in g]
            return []
        tenant_groups_res = []
        member_groups_res = []
        for group in tenant_groups:
            tenant_groups_res.extend(group_list(group))
        for group in member_groups:
            member_groups_res.extend(group_list(group))
        return tenant_groups_res, member_groups_res

    AccessGroup = apps.get_model('accrete', 'AccessGroup')
    data = {}
    flat_tenant_groups, flat_member_groups = _flat_groups()
    if flat_tenant_groups:
        data.update(tenant_groups=[])
        access_groups = AccessGroup.objects.filter(
            code__in=flat_tenant_groups,
            apply_on='tenant'
        ).all()
        group_data = {item[0]: item[1] for item in access_groups.values_list('code', 'name')}
        for group in tenant_groups:
            if isinstance(group, tuple):
                data['tenant_groups'].append(' & '.join([group_data.get(g, g) for g in group]))
            else:
                data['tenant_groups'].append(group_data.get(group, group))
    if flat_member_groups:
        data.update(member_groups=[])
        access_groups = AccessGroup.objects.filter(
            code__in=flat_member_groups,
            apply_on='member'
        ).all()
        group_data = {item[0]: item[1] for item in access_groups.values_list('code', 'name')}
        for group in member_groups:
            if isinstance(group, tuple):
                data['member_groups'].append(' & '.join([group_data.get(g, g) for g in group]))
            else:
                data['member_groups'].append(group_data.get(group, group))
    return data


def access_denied_page_response(request, group_data):
    return ui.WindowResponse(
        title=str(_('Access Denied')),
        overview_template='mirox/base/group_not_set.html',
        context=dict(groups=group_data),
        is_centered=True
    ).response(request)


def access_denied_modal_response(request, group_data):
    res = ui.ModalResponse(
        template='mirox/base/group_not_set_modal.html',
        title=str(_('Access Denied')),
        modal_id='group-missing-modal',
        context=dict(groups=group_data)
    ).response(request)
    res.headers['HX-Reswap'] = 'none'
    res.headers['HX-Push-Url'] = 'false'
    return res



# Generated by Django 5.2.8 on 2026-02-11 10:03

import django.core.validators
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('sequence', '0004_sequence_has_subsequence_per_year_sequence_padding_and_more'),
    ]

    operations = [
        migrations.AlterField(
            model_name='sequence',
            name='padding',
            field=models.IntegerField(default=0, help_text='Pads the sequence with zeros to the given length.', validators=[django.core.validators.MaxValueValidator(10)], verbose_name='Padding'),
        ),
    ]



from decimal import Decimal
from datetime import date, datetime

from django.conf import settings
from django.db.models.functions import Lower
from django.utils.translation import gettext_lazy as _
from django.db import models
from accrete.models import TranslatedCharField


class Activity(models.Model):

    class Meta:
        verbose_name = _('Activity')
        verbose_name_plural = _('Activities')
        db_table = 'accrete_activity'

    code = models.CharField(
        verbose_name=_('Code'),
        default='',
        blank=True
    )

    tenant = models.ForeignKey(
        verbose_name=_('Tenant'),
        to='accrete.Tenant',
        related_name='accrete_activities',
        on_delete=models.CASCADE,
        null=True,
        blank=True
    )

    user = models.ForeignKey(
        to=settings.AUTH_USER_MODEL,
        related_name='accrete_activities',
        on_delete=models.SET_NULL,
        null=True,
        blank=True
    )

    description = TranslatedCharField(
        verbose_name=_('Description'),
        null=True,
        blank=True
    )

    model = models.CharField(
        verbose_name=_('Model'),
        max_length=255,
        null=True,
        blank=True,
        help_text=(
            'Information about the triggering model, log lines can have deviating models'
            'Combination of app label and model name separated by a dot.'
        ),
    )

    object_id = models.BigIntegerField(
        verbose_name='Object ID',
        null=True,
        blank=True
    )

    log_date = models.DateTimeField(
        verbose_name=_('Date'),
        auto_now_add=True
    )

    def __str__(self):
        return f'{self.code}({self.pk})'


class Log(models.Model):

    class Meta:
        verbose_name = _('Log')
        verbose_name_plural = _('Logs')
        db_table = 'accrete_log'
        indexes = [
            models.Index(fields=['model', 'object_id', 'tenant_id'])
        ]
        constraints = [
            models.UniqueConstraint(
                name='unique_group_fields',
                fields=('model', 'field', 'object_id', 'log_date')
            )
        ]

    tenant = models.ForeignKey(
        verbose_name=_('Tenant'),
        to='accrete.Tenant',
        on_delete=models.CASCADE,
        null=True,
        blank=True
    )

    model = models.CharField(
        verbose_name=_('Model'),
        max_length=255,
        help_text='Combination of app label and model name separated by a dot'
    )

    field = models.CharField(
        verbose_name=_('Field'),
        max_length=255
    )

    object_id = models.BigIntegerField(
        verbose_name='Object ID'
    )

    log_date = models.DateTimeField(
        verbose_name=_('Date'),
        auto_now_add=True
    )

    old_value_type = models.CharField(
        verbose_name=_('Old Value Type'),
        choices=[
            ('fk', 'Foreign Key'),
            ('int', 'Integer'),
            ('float', 'Float'),
            ('decimal', 'Decimal'),
            ('bool', 'Boolean'),
            ('str', 'String'),
            ('date', 'Date'),
            ('datetime', 'Date Time'),
            ('json', 'JSON')
        ],
        max_length=100,
    )

    new_value_type = models.CharField(
        verbose_name=_('New Value Type'),
        choices=[
            ('fk', 'Foreign Key'),
            ('int', 'Integer'),
            ('float', 'Float'),
            ('decimal', 'Decimal'),
            ('bool', 'Boolean'),
            ('str', 'String'),
            ('date', 'Date'),
            ('datetime', 'Date Time'),
            ('json', 'JSON')
        ],
        max_length=100
    )

    old_value = models.TextField(
        verbose_name=_('Old Value'),
        null=True,
        blank=True
    )

    new_value = models.TextField(
        verbose_name=_('New Value'),
        null=True,
        blank=True
    )

    user = models.ForeignKey(
        to=settings.AUTH_USER_MODEL,
        related_name='logs',
        on_delete=models.SET_NULL,
        null=True,
        blank=True
    )

    activity = models.ForeignKey(
        to='log.Activity',
        related_name='logs',
        on_delete=models.CASCADE,
        blank=True,
        null=True
    )

    def __str__(self):
        return f'{self.model}.{self.field},{self.object_id}'

    def cast_value(self):
        if self.new_value is None:
            return None
        if self.new_value_type == 'fk':
            return self.cast_fk()
        if self.new_value_type == 'int':
            return self.cast_int()
        if self.new_value_type == 'float':
            return self.cast_float()
        if self.new_value_type == 'decimal':
            return self.cast_decimal()
        if self.new_value_type == 'bool':
            return self.cast_bool()
        if self.new_value_type == 'str':
            return self.cast_str()
        if self.new_value_type == 'date':
            return self.cast_date()
        if self.new_value_type == 'datetime':
            return self.cast_date_time()
        if self.new_value_type == 'json':
            return self.cast_json()

    def cast_fk(self):
        return self.cast_int()

    def cast_int(self):
        if self.new_value == '':
            return None
        return int(self.new_value)

    def cast_float(self):
        if self.new_value == '':
            return 0.0
        return float(self.new_value)

    def cast_decimal(self):
        if self.new_value == '':
            return Decimal(0)
        return Decimal(self.new_value)

    def cast_bool(self):
        return self.new_value == 'True'

    def cast_str(self):
        return self.new_value

    def cast_date(self):
        if self.new_value == '':
            return None
        return date.fromisoformat(str(self.new_value))

    def cast_date_time(self):
        if self.new_value == '':
            return None
        return datetime.fromisoformat(self.new_value)
    
    def cast_json(self):
        return self.new_value


class LogConfig(models.Model):

    class Meta:
        verbose_name = _('Log Configuration')
        verbose_name_plural = _('Log Configs')
        db_table = 'accrete_log_config'
        ordering = [Lower('model')]
        indexes = [
            models.Index(fields=['model'])
        ]
        constraints = [
            models.UniqueConstraint(
                name='unique_model',
                fields=('model',)
            )
        ]

    model = models.CharField(
        verbose_name=_('Model'),
        max_length=255,
        help_text=_('Combination of app label and model name separated by a dot')
    )

    ignore_errors = models.BooleanField(
        verbose_name=_('Ignore Errors'),
        default=False,
        help_text=_('If set, exceptions during log creation will be ignored')
    )

    exclude_fields = models.BooleanField(
        verbose_name=_('Exclude Fields'),
        default=False,
        help_text=_(
            'If set, Log Configuration Fields will be excluded from logging. '
            'Otherwise only configured fields will be logged.'
        )
    )

    def __str__(self):
        return self.model


class LogConfigField(models.Model):

    class Meta:
        verbose_name = _('Log Configuration Field')
        verbose_name_plural = _('Log Configuration Fields')
        db_table = 'accrete_log_config_field'
        ordering = [Lower('field_name')]
        indexes = [
            models.Index(fields=['log_config'])
        ]
        constraints = [
            models.UniqueConstraint(
                name='unique_config_field_name',
                fields=('log_config', 'field_name')
            )
        ]

    log_config = models.ForeignKey(
        verbose_name=_('Log Configuration'),
        to='log.LogConfig',
        related_name='fields',
        on_delete=models.CASCADE
    )

    field_name = models.CharField(
        verbose_name=_('Field Name'),
        max_length=255,
        help_text=_('Name of the field to log')
    )

    def __str__(self):
        return f'{self.log_config}.{self.field_name}'



# Generated by Django 5.2.7 on 2025-10-06 18:59

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('log', '0006_remove_log_accrete_log_model_cf888c_idx_and_more'),
    ]

    operations = [
        migrations.AlterField(
            model_name='log',
            name='new_value_type',
            field=models.CharField(choices=[('fk', 'Foreign Key'), ('int', 'Integer'), ('float', 'Float'), ('decimal', 'Decimal'), ('bool', 'Boolean'), ('str', 'String'), ('date', 'Date'), ('datetime', 'Date Time'), ('json', 'JSON')], max_length=100, verbose_name='New Value Type'),
        ),
        migrations.AlterField(
            model_name='log',
            name='old_value_type',
            field=models.CharField(choices=[('fk', 'Foreign Key'), ('int', 'Integer'), ('float', 'Float'), ('decimal', 'Decimal'), ('bool', 'Boolean'), ('str', 'String'), ('date', 'Date'), ('datetime', 'Date Time'), ('json', 'JSON')], max_length=100, verbose_name='Old Value Type'),
        ),
    ]



from django.contrib.auth import get_user_model
from django.contrib.auth.backends import ModelBackend

User = get_user_model()


class LoginModelBackend(ModelBackend):
    """
    This is a ModelBacked that allows authentication with the login field.
    """

    def authenticate(self, request, username=None, password=None, **kwargs):
        if username is None:
            return None
        try:
            user = User.objects.get(login=username)
        except User.DoesNotExist:
            return None
        if user.check_password(password):
            return user
        return None
